# sage_setup: distribution = sagemath-categories
r"""
Scheme morphism

.. NOTE::

    You should never create the morphisms directly. Instead, use the
    :meth:`~sage.schemes.generic.scheme.hom` and
    :meth:`~sage.structure.parent.Hom` methods that are inherited by
    all schemes.

If you want to extend the Sage library with some new kind of scheme,
your new class (say, ``MyScheme``) should provide a method

* ``MyScheme._morphism(*args, **kwds)`` returning a morphism
  between two schemes in your category, usually defined via
  polynomials. Your morphism class should derive from
  :class:`SchemeMorphism_polynomial`. These morphisms will usually be
  elements of the Hom-set
  :class:`~sage.schemes.generic.homset.SchemeHomset_generic`.

Optionally, you can also provide a special Hom-set class for your
subcategory of schemes. If you want to do this, you should also
provide a method

* ``MyScheme._homset(*args, **kwds)`` returning a
  Hom-set, which must be an element of a derived class of
  :class:`~sage.schemes.generic.homset.SchemeHomset_generic`. If your
  new Hom-set class does not use ``MyScheme._morphism`` then you
  do not have to provide it.

Note that points on schemes are morphisms `\mathrm{Spec}(K)\to X`, too. But we
typically use a different notation, so they are implemented in a different
derived class. For this, you should implement a method

* ``MyScheme._point(*args, **kwds)`` returning a point, that is, a morphism
  `\mathrm{Spec}(K)\to X`. Your point class should derive from
  :class:`SchemeMorphism_point`.

Optionally, you can also provide a special Hom-set for the points, for
example the point Hom-set can provide a method to enumerate all
points. If you want to do this, you should also provide a method

* ``MyScheme._point_homset(*args, **kwds)`` returning
  the :mod:`~sage.schemes.generic.homset` of points. The Hom-sets of
  points are implemented in classes named ``SchemeHomset_points_...``.
  If your new Hom-set class does not use ``MyScheme._point`` then
  you do not have to provide it.

AUTHORS:

- David Kohel, William Stein

- William Stein (2006-02-11): fixed bug where P(0,0,0) was allowed as
  a projective point.

- Volker Braun (2011-08-08): Renamed classes, more documentation, misc
  cleanups.

- Ben Hutz (June 2012): added support for projective ring

- Simon King (2013-10): copy the changes of :class:`~sage.categories.morphism.Morphism`
  that have been introduced in :issue:`14711`.
"""

# ****************************************************************************
#       Copyright (C) 2013 Simon King <simon.king@uni-jena.de>
#       Copyright (C) 2011 Volker Braun <vbraun.name@gmail.com>
#       Copyright (C) 2006 David Kohel <kohel@maths.usyd.edu.au>
#       Copyright (C) 2006 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

import operator

from sage.arith.power import generic_power
from sage.categories.homset import Homset, Hom, End
from sage.categories.map import FormalCompositeMap, Map
from sage.categories.morphism import SetMorphism
from sage.misc.constant_function import ConstantFunction
from sage.misc.lazy_attribute import lazy_attribute
from sage.rings.fraction_field import FractionField_generic
from sage.rings.fraction_field_element import FractionFieldElement
from sage.schemes.generic.algebraic_scheme import AlgebraicScheme_subscheme
from sage.structure.element import Element, parent, coercion_model
from sage.structure.richcmp import richcmp
from sage.structure.sequence import Sequence


def is_SchemeMorphism(f):
    """
    Test whether ``f`` is a scheme morphism.

    INPUT:

    - ``f`` -- anything

    OUTPUT:

    boolean; return ``True`` if ``f`` is a scheme morphism or a point
    on an elliptic curve.

    EXAMPLES::

        sage: A.<x,y> = AffineSpace(QQ, 2); H = A.Hom(A)
        sage: f = H([y, x^2 + y]); f
        Scheme endomorphism of Affine Space of dimension 2 over Rational Field
          Defn: Defined on coordinates by sending (x, y) to (y, x^2 + y)
        sage: from sage.schemes.generic.morphism import is_SchemeMorphism
        sage: is_SchemeMorphism(f)
        doctest:warning...
        DeprecationWarning: The function is_SchemeMorphism is deprecated;
        use 'isinstance(..., SchemeMorphism)' instead.
        See https://github.com/sagemath/sage/issues/38296 for details.
        True
    """
    from sage.misc.superseded import deprecation
    deprecation(38296,
                "The function is_SchemeMorphism is deprecated; "
                "use 'isinstance(..., SchemeMorphism)' instead.")
    return isinstance(f, SchemeMorphism)


class SchemeMorphism(Element):
    """
    Base class for scheme morphisms.

    INPUT:

    - ``parent`` -- the parent of the morphism

    .. TODO::

        For historical reasons, :class:`SchemeMorphism` copies code from
        :class:`~sage.categories.map.Map` rather than inheriting from it.
        Proper inheritance should be used instead. See :issue:`14711`.

    EXAMPLES::

        sage: X = Spec(ZZ)
        sage: Hom = X.Hom(X)
        sage: from sage.schemes.generic.morphism import SchemeMorphism
        sage: f = SchemeMorphism(Hom)
        sage: type(f)
        <class 'sage.schemes.generic.morphism.SchemeMorphism'>

    TESTS::

        sage: A2 = AffineSpace(QQ, 2)
        sage: A2.structure_morphism().domain()
        Affine Space of dimension 2 over Rational Field
        sage: A2.structure_morphism().category()
        Category of homsets of schemes
    """

    def __init__(self, parent, codomain=None):
        """
        The Python constructor.

        EXAMPLES::

            sage: X = Spec(ZZ)
            sage: Hom = X.Hom(X)
            sage: from sage.schemes.generic.morphism import SchemeMorphism
            sage: f = SchemeMorphism(Hom)
            sage: type(f)
            <class 'sage.schemes.generic.morphism.SchemeMorphism'>
        """
        if codomain is not None:
            parent = Hom(parent, codomain)
        if not isinstance(parent, Homset):
            raise TypeError("parent (=%s) must be a Homspace" % parent)
        Element.__init__(self, parent)
        self._codomain = parent.codomain()

    @lazy_attribute
    def domain(self):
        r"""
        The constant function from the domain.

        EXAMPLES::

            sage: A.<x,y> = AffineSpace(QQ['x,y'])
            sage: H = A.Hom(A)
            sage: f = H([y, x^2 + y])
            sage: f.domain() is A
            True
        """
        return ConstantFunction(self.parent().domain())

    @lazy_attribute
    def codomain(self):
        r"""
        The constant function from the codomain.

        EXAMPLES::

            sage: A.<x,y> = AffineSpace(QQ['x,y'])
            sage: H = A.Hom(A)
            sage: f = H([y, x^2 + y])
            sage: f.codomain() is A
            True
        """
        return ConstantFunction(self._codomain)

    # We copy methods of sage.categories.map.Map, to make
    # a future transition of SchemeMorphism to a sub-class of Morphism
    # easier.
    def __call__(self, x, *args, **kwds):
        """
        Do not override this method!

        For implementing application of maps, implement a method
        ``_call_(self, x)`` and/or a method ``_call_with_args(x, args, kwds)`.
        In these methods, you can assume that ``x`` belongs to the domain of
        this morphism, ``args`` is a tuple and ``kwds`` is a dict.

        EXAMPLES::

            sage: R.<x,y> = QQ[]
            sage: A.<x,y> = AffineSpace(R)
            sage: H = A.Hom(A)
            sage: f = H([y, x^2 + y])
            sage: f([2,3])    # indirect doctest
            (3, 7)

        An example with optional arguments::

            sage: PS.<x,y> = ProjectiveSpace(QQ, 1)
            sage: H = Hom(PS, PS)
            sage: f = H([x^3, x*y^2])
            sage: P = PS(0, 1)
            sage: f(P, check=False)     # indirect doctest
            (0 : 0)
        """
        P = parent(x)
        D = self.domain()
        if P is D: # we certainly want to call _call_/with_args
            if not args and not kwds:
                return self._call_(x)
            return self._call_with_args(x, args, kwds)
        # Is there coercion?
        converter = D._internal_coerce_map_from(P)
        if converter is None:
            try:
                return self.pushforward(x, *args, **kwds)
            except (AttributeError, TypeError, NotImplementedError):
                pass # raise TypeError, "%s must be coercible into %s"%(x, self.domain())
            # Here, we would like to do
            ##try:
            ##    x = D(x).
            ##except (TypeError, NotImplementedError):
            ##    raise TypeError, "%s fails to convert into the map's domain %s, but a `pushforward` method is not properly implemented"%(x, self.domain())
            # However, this would involve a test whether x.codomain() ==
            # self. This would trigger a Groebner basis computation, that
            # (1) could be slow and (2) could involve an even slower toy
            # implementation, resulting in a warning.
            #
            # Contract: If x is a scheme morphism point, then _call_ knows
            # what to do with it (e.g., use the _coords attribute). Otherwise,
            # we can try a conversion into the domain (e.g., if x is a list),
            # WITHOUT to trigger a Groebner basis computation.
            if kwds.get('check', True):
                if not isinstance(x, SchemeMorphism_point):
                    try:
                        x = D(x)
                    except (TypeError, NotImplementedError):
                        raise TypeError("%s fails to convert into the map's domain %s, but a `pushforward` method is not properly implemented" % (x, self.domain()))
                elif self.domain() != x.codomain():
                    raise TypeError("%s fails to convert into the map's domain %s, but a `pushforward` method is not properly implemented" % (x, self.domain()))
        else:
            x = converter(x)
        if not args and not kwds:
            return self._call_(x)
        return self._call_with_args(x, args, kwds)

    def _repr_defn(self):
        r"""
        Return a string representation of the definition of ``self``.

        OUTPUT: string

        EXAMPLES::

            sage: X = Spec(ZZ)
            sage: Hom = X.Hom(X)
            sage: from sage.schemes.generic.morphism import SchemeMorphism
            sage: f = SchemeMorphism(Hom)
            sage: f._repr_defn()
            Traceback (most recent call last):
            ...
            NotImplementedError
        """
        raise NotImplementedError

    def _repr_type(self):
        r"""
        Return a string representation of the type of ``self``.

        OUTPUT: string

        EXAMPLES::

            sage: A2 = AffineSpace(QQ, 2)
            sage: A2.structure_morphism()  # indirect doctest
            Scheme morphism:
              From: Affine Space of dimension 2 over Rational Field
              To:   Spectrum of Rational Field
              Defn: Structure map
        """
        return "Scheme"

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        OUTPUT: string

        EXAMPLES::

            sage: X = Spec(ZZ)
            sage: Hom = X.Hom(X)
            sage: from sage.schemes.generic.morphism import SchemeMorphism
            sage: f = SchemeMorphism(Hom)
            sage: f._repr_()
            Traceback (most recent call last):
            ...
            NotImplementedError
        """
        if self.is_endomorphism():
            s = "%s endomorphism of %s" % (self._repr_type(), self.domain())
        else:
            s = "%s morphism:" % self._repr_type()
            s += "\n  From: %s" % self.domain()
            s += "\n  To:   %s" % self._codomain
        d = self._repr_defn()
        if d != '':
            s += "\n  Defn: %s" % ('\n        '.join(self._repr_defn().split('\n')))
        return s

    def __mul__(self, right):
        """
        We can currently only multiply scheme morphisms.

        If one factor is an identity morphism, the other is returned.
        Otherwise, a formal composition of maps obtained from the scheme
        morphisms is returned.

        EXAMPLES:

        Identity maps do not contribute to the product::

            sage: X = AffineSpace(QQ, 2)
            sage: id = X.identity_morphism()
            sage: id^0    # indirect doctest
            Scheme endomorphism of Affine Space of dimension 2 over Rational Field
              Defn: Identity map
            sage: id^2
            Scheme endomorphism of Affine Space of dimension 2 over Rational Field
              Defn: Identity map

        Here, we see a formal composition::

            sage: X = AffineSpace(QQ, 2)
            sage: f = X.structure_morphism()
            sage: Y = Spec(QQ)
            sage: g = Y.structure_morphism()
            sage: g * f    # indirect doctest
            Composite map:
              From: Affine Space of dimension 2 over Rational Field
              To:   Spectrum of Integer Ring
              Defn:   Generic morphism:
                      From: Affine Space of dimension 2 over Rational Field
                      To:   Spectrum of Rational Field
                    then
                      Generic morphism:
                      From: Spectrum of Rational Field
                      To:   Spectrum of Integer Ring

        Of course, the codomain of the first factor must coincide with the
        domain of the second factor::

            sage: f * g
            Traceback (most recent call last):
            ...
            TypeError: self (=Scheme morphism:
              From: Affine Space of dimension 2 over Rational Field
              To:   Spectrum of Rational Field
              Defn: Structure map) domain must equal right (=Scheme morphism:
              From: Spectrum of Rational Field
              To:   Spectrum of Integer Ring
              Defn: Structure map) codomain
        """
        if not isinstance(right, SchemeMorphism):
            return coercion_model.bin_op(self, right, operator.mul)
        if right.codomain() != self.domain():
            raise TypeError("self (=%s) domain must equal right (=%s) codomain" % (self, right))
        if isinstance(self, SchemeMorphism_id):
            return right
        if isinstance(right, SchemeMorphism_id):
            return self
        return self._composition(right)

    def __pow__(self, n, dummy=None):
        """
        Exponentiate an endomorphism.

        INPUT:

        - ``n`` -- integer; the exponent

        OUTPUT: a composite map that belongs to the same endomorphism set as ``self``

        EXAMPLES::

            sage: X = AffineSpace(QQ, 2)
            sage: id = X.identity_morphism()
            sage: id^0
            Scheme endomorphism of Affine Space of dimension 2 over Rational Field
              Defn: Identity map
            sage: id^2
            Scheme endomorphism of Affine Space of dimension 2 over Rational Field
              Defn: Identity map
        """
        if not self.is_endomorphism():
            raise TypeError("self must be an endomorphism.")
        if n == 0:
            return self.domain().identity_morphism()
        return generic_power(self, n)

    def category(self):
        """
        Return the category of the Hom-set.

        OUTPUT: a category

        EXAMPLES::

            sage: A2 = AffineSpace(QQ, 2)
            sage: A2.structure_morphism().category()
            Category of homsets of schemes
        """
        return self.parent().category()

    def category_for(self):
        """
        Return the category which this morphism belongs to.

        EXAMPLES::

            sage: A2 = AffineSpace(QQ, 2)
            sage: A2.structure_morphism().category_for()
            Category of schemes
        """
        return self.parent().homset_category()

    def is_endomorphism(self) -> bool:
        """
        Return whether the morphism is an endomorphism.

        OUTPUT: boolean; whether the domain and codomain are identical

        EXAMPLES::

            sage: X = AffineSpace(QQ, 2)
            sage: X.structure_morphism().is_endomorphism()
            False
            sage: X.identity_morphism().is_endomorphism()
            True
        """
        return self.parent().is_endomorphism_set()

    def base_ring(self):
        r"""
        Return the base ring of ``self``, that is, the ring over which
        the defining polynomials of ``self`` are defined.

        OUTPUT: ring

        EXAMPLES::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: H = Hom(P, P)
            sage: f = H([3/5*x^2, 6*y^2])
            sage: f.base_ring()
            Rational Field

        ::

            sage: R.<t> = PolynomialRing(ZZ, 1)
            sage: P.<x,y> = ProjectiveSpace(R, 1)
            sage: H = Hom(P, P)
            sage: f = H([3*x^2, y^2])
            sage: f.base_ring()
            Multivariate Polynomial Ring in t over Integer Ring

        Points have correct base rings too (:issue:`34336`)::

            sage: x = P(t, 5); x
            (t : 5)
            sage: x.base_ring()
            Multivariate Polynomial Ring in t over Integer Ring

        ::

            sage: # needs sage.rings.finite_rings sage.schemes
            sage: E = EllipticCurve(GF((17,2)), [1,2,3,4,5])
            sage: P = E.random_point()
            sage: P.base_ring()
            Finite Field in z2 of size 17^2
        """
        return self.domain().base_ring()

    def _composition(self, right):
        """
        A helper for multiplying maps by composition.

        .. WARNING::

            Do not override this method! Override :meth:`_composition_`
            instead.

        EXAMPLES::

            sage: X = AffineSpace(QQ, 2)
            sage: f = X.structure_morphism()
            sage: Y = Spec(QQ)
            sage: g = Y.structure_morphism()
            sage: g * f    # indirect doctest
             Composite map:
              From: Affine Space of dimension 2 over Rational Field
              To:   Spectrum of Integer Ring
              Defn:   Generic morphism:
                      From: Affine Space of dimension 2 over Rational Field
                      To:   Spectrum of Rational Field
                    then
                      Generic morphism:
                      From: Spectrum of Rational Field
                      To:   Spectrum of Integer Ring

            sage: f * g
            Traceback (most recent call last):
            ...
            TypeError: self (=Scheme morphism:
              From: Affine Space of dimension 2 over Rational Field
              To:   Spectrum of Rational Field
              Defn: Structure map) domain
            must equal right (=Scheme morphism:
              From: Spectrum of Rational Field
              To:   Spectrum of Integer Ring
              Defn: Structure map) codomain
        """
        category = self.category_for()._meet_(right.category_for())
        H = Hom(right.domain(), self._codomain, category)
        return self._composition_(right, H)

    def _composition_(self, right, homset):
        """
        Helper to construct the composition of two morphisms.

        Override this if you want to have a different behaviour of composition

        INPUT:

        - ``right`` -- a map or callable
        - ``homset`` -- a homset containing the composed map

        OUTPUT:

        An element of ``homset``. The output is obtained by converting the
        arguments to :class:`~sage.categories.morphism.SetMorphism` if
        necessary, and then forming a :class:`~sage.categories.map.FormalCompositeMap`

        EXAMPLES::

            sage: X = AffineSpace(QQ, 2)
            sage: f = X.structure_morphism()
            sage: Y = Spec(QQ)
            sage: g = Y.structure_morphism()
            sage: g * f    # indirect doctest
            Composite map:
              From: Affine Space of dimension 2 over Rational Field
              To:   Spectrum of Integer Ring
              Defn:   Generic morphism:
                      From: Affine Space of dimension 2 over Rational Field
                      To:   Spectrum of Rational Field
                    then
                      Generic morphism:
                      From: Spectrum of Rational Field
                      To:   Spectrum of Integer Ring
        """
        if not isinstance(right, Map):
            right = SetMorphism(right.parent(), right)
        return FormalCompositeMap(homset, right, SetMorphism(self.parent(),self))

    def glue_along_domains(self, other):
        r"""
        Glue two morphisms.

        INPUT:

        - ``other`` -- a scheme morphism with the same domain

        OUTPUT:

        Assuming that ``self`` and ``other`` are open immersions with the same
        domain, return scheme obtained by gluing along the images.

        EXAMPLES:

        We construct a scheme isomorphic to the projective line over
        `\mathrm{Spec}(\QQ)` by gluing two copies of `\mathbb{A}^1`
        minus a point::

            sage: # needs sage.libs.singular
            sage: R.<x,y> = PolynomialRing(QQ, 2)
            sage: S.<xbar, ybar> = R.quotient(x*y - 1)
            sage: Rx = PolynomialRing(QQ, 'x')
            sage: i1 = Rx.hom([xbar])
            sage: Ry = PolynomialRing(QQ, 'y')
            sage: i2 = Ry.hom([ybar])
            sage: Sch = Schemes()
            sage: f1 = Sch(i1)
            sage: f2 = Sch(i2)

        Now f1 and f2 have the same domain, which is a
        `\mathbb{A}^1` minus a point. We glue along the domain::

            sage: # needs sage.libs.singular
            sage: P1 = f1.glue_along_domains(f2); P1
            Scheme obtained by gluing X and Y along U, where
              X: Spectrum of Univariate Polynomial Ring in x over Rational Field
              Y: Spectrum of Univariate Polynomial Ring in y over Rational Field
              U: Spectrum of Quotient of Multivariate Polynomial Ring in x, y
                 over Rational Field by the ideal (x*y - 1)
            sage: a, b = P1.gluing_maps()
            sage: a
            Affine Scheme morphism:
              From: Spectrum of Quotient of Multivariate Polynomial Ring in x, y
                    over Rational Field by the ideal (x*y - 1)
              To:   Spectrum of Univariate Polynomial Ring in x over Rational Field
              Defn: Ring morphism:
                      From: Univariate Polynomial Ring in x over Rational Field
                      To:   Quotient of Multivariate Polynomial Ring in x, y over
                            Rational Field by the ideal (x*y - 1)
                      Defn: x |--> xbar
            sage: b
            Affine Scheme morphism:
              From: Spectrum of Quotient of Multivariate Polynomial Ring in x, y
                    over Rational Field by the ideal (x*y - 1)
              To:   Spectrum of Univariate Polynomial Ring in y over Rational Field
              Defn: Ring morphism:
                      From: Univariate Polynomial Ring in y over Rational Field
                      To:   Quotient of Multivariate Polynomial Ring in x, y over
                            Rational Field by the ideal (x*y - 1)
                      Defn: y |--> ybar
        """
        from . import glue
        return glue.GluedScheme(self, other)


class SchemeMorphism_id(SchemeMorphism):
    """
    Return the identity morphism from `X` to itself.

    INPUT:

    - ``X`` -- the scheme

    EXAMPLES::

        sage: X = Spec(ZZ)
        sage: X.identity_morphism()  # indirect doctest
        Scheme endomorphism of Spectrum of Integer Ring
          Defn: Identity map
    """
    def __init__(self, X):
        """
        The Python constructor.

        See :class:`SchemeMorphism_id` for details.

        TESTS::

            sage: Spec(ZZ).identity_morphism()
            Scheme endomorphism of Spectrum of Integer Ring
              Defn: Identity map
        """
        SchemeMorphism.__init__(self, X.Hom(X))

    def _repr_defn(self):
        r"""
        Return a string representation of the definition of ``self``.

        OUTPUT: string

        EXAMPLES::

            sage: Spec(ZZ).identity_morphism()._repr_defn()
            'Identity map'
        """
        return 'Identity map'


class SchemeMorphism_structure_map(SchemeMorphism):
    r"""
    The structure morphism.

    INPUT:

    - ``parent`` -- Hom-set with codomain equal to the base scheme of
      the domain

    EXAMPLES::

        sage: Spec(ZZ).structure_morphism()    # indirect doctest
        Scheme endomorphism of Spectrum of Integer Ring
          Defn: Structure map
    """
    def __init__(self, parent, codomain=None):
        """
        The Python constructor.

        See :class:`SchemeMorphism_structure_map` for details.

        TESTS::

            sage: from sage.schemes.generic.morphism import SchemeMorphism_structure_map
            sage: SchemeMorphism_structure_map( Spec(QQ).Hom(Spec(ZZ)) )
            Scheme morphism:
              From: Spectrum of Rational Field
              To:   Spectrum of Integer Ring
              Defn: Structure map
        """
        SchemeMorphism.__init__(self, parent, codomain=None)
        if self.domain().base_scheme() != self._codomain:
            raise ValueError("parent must have codomain equal the base scheme of domain.")

    def _repr_defn(self):
        r"""
        Return a string representation of the definition of ``self``.

        OUTPUT: string

        EXAMPLES::

            sage: Spec(ZZ).structure_morphism()._repr_defn()
            'Structure map'
        """
        return 'Structure map'


class SchemeMorphism_spec(SchemeMorphism):
    """
    Morphism of spectra of rings.

    INPUT:

    - ``parent`` -- Hom-set whose domain and codomain are affine schemes

    - ``phi`` -- a ring morphism with matching domain and codomain

    - ``check`` -- boolean (default: ``True``); whether to
      check the input for consistency

    EXAMPLES::

        sage: R.<x> = PolynomialRing(QQ)
        sage: phi = R.hom([QQ(7)]); phi
        Ring morphism:
          From: Univariate Polynomial Ring in x over Rational Field
          To:   Rational Field
          Defn: x |--> 7

        sage: X = Spec(QQ); Y = Spec(R)
        sage: f = X.hom(phi); f
        Affine Scheme morphism:
          From: Spectrum of Rational Field
          To:   Spectrum of Univariate Polynomial Ring in x over Rational Field
          Defn: Ring morphism:
                  From: Univariate Polynomial Ring in x over Rational Field
                  To:   Rational Field
                  Defn: x |--> 7

        sage: f.ring_homomorphism()
        Ring morphism:
          From: Univariate Polynomial Ring in x over Rational Field
          To:   Rational Field
          Defn: x |--> 7
    """
    def __init__(self, parent, phi, check=True):
        """
        The Python constructor.

        See :class:`SchemeMorphism_structure_map` for details.

        TESTS::

            sage: from sage.schemes.generic.morphism import SchemeMorphism_spec
            sage: SchemeMorphism_spec(Spec(QQ).Hom(Spec(ZZ)), ZZ.hom(QQ))
            Affine Scheme morphism:
              From: Spectrum of Rational Field
              To:   Spectrum of Integer Ring
              Defn: Natural morphism:
                      From: Integer Ring
                      To:   Rational Field
        """
        SchemeMorphism.__init__(self, parent)
        if check:
            from sage.categories.rings import Rings
            if not (isinstance(phi, Map) and phi.category_for().is_subcategory(Rings())):
                raise TypeError("phi (=%s) must be a ring homomorphism" % phi)
            if phi.domain() != parent.codomain().coordinate_ring():
                raise TypeError("phi (=%s) must have domain %s"
                                % (phi, parent.codomain().coordinate_ring()))
            if phi.codomain() != parent.domain().coordinate_ring():
                raise TypeError("phi (=%s) must have codomain %s"
                                % (phi, parent.domain().coordinate_ring()))
        self.__ring_homomorphism = phi

    def _call_(self, x):
        r"""
        Make morphisms callable.

        INPUT:

        - ``x`` -- a scheme point

        OUTPUT: the image scheme point

        EXAMPLES:

        The following fails because inverse images of prime ideals
        under ring homomorphisms are not yet implemented::

            sage: R.<x> = PolynomialRing(QQ)
            sage: phi = R.hom([QQ(7)])
            sage: X = Spec(QQ); Y = Spec(R)
            sage: f = X.hom(phi)
            sage: f(X.an_element())    # indirect doctest
            Traceback (most recent call last):
            ...
            NotImplementedError: inverse not implemented for morphisms of
            Rational Field
        """
        # By virtue of argument preprocessing in __call__, we can assume that
        # x is a topological scheme point of self
        S = self.ring_homomorphism().inverse_image(x.prime_ideal())
        return self._codomain(S)

    def _repr_type(self):
        r"""
        Return a string representation of the type of ``self``.

        OUTPUT: string

        EXAMPLES::

            sage: R.<x> = PolynomialRing(QQ)
            sage: phi = R.hom([QQ(7)])
            sage: X = Spec(QQ); Y = Spec(R)
            sage: f = X.hom(phi)
            sage: f._repr_type()
            'Affine Scheme'
        """
        return "Affine Scheme"

    def _repr_defn(self):
        r"""
        Return a string representation of the definition of ``self``.

        OUTPUT: string

        EXAMPLES::

            sage: R.<x> = PolynomialRing(QQ)
            sage: phi = R.hom([QQ(7)])
            sage: X = Spec(QQ); Y = Spec(R)
            sage: f = X.hom(phi)
            sage: print(f._repr_defn())
            Ring morphism:
              From: Univariate Polynomial Ring in x over Rational Field
              To:   Rational Field
              Defn: x |--> 7
        """
        return repr(self.ring_homomorphism())

    def ring_homomorphism(self):
        """
        Return the underlying ring homomorphism.

        OUTPUT: a ring homomorphism

        EXAMPLES::

            sage: R.<x> = PolynomialRing(QQ)
            sage: phi = R.hom([QQ(7)])
            sage: X = Spec(QQ); Y = Spec(R)
            sage: f = X.hom(phi)
            sage: f.ring_homomorphism()
            Ring morphism:
              From: Univariate Polynomial Ring in x over Rational Field
              To:   Rational Field
              Defn: x |--> 7
        """
        return self.__ring_homomorphism


############################################################################
# Morphisms between schemes given on points. The _affine and _projective below
# refer to the CODOMAIN. The domain can be either affine or projective
# regardless of the class
############################################################################

class SchemeMorphism_polynomial(SchemeMorphism):
    r"""
    A morphism of schemes determined by polynomials that define what
    the morphism does on points in the ambient space.

    INPUT:

    - ``parent`` -- Hom-set whose domain and codomain are affine or
      projective schemes

    - ``polys`` -- list/tuple/iterable of polynomials defining the
      scheme morphism

    - ``check`` -- boolean (default: ``True``); whether to
      check the input for consistency

    EXAMPLES:

    An example involving the affine plane::

        sage: R.<x,y> = QQ[]
        sage: A2 = AffineSpace(R)
        sage: H = A2.Hom(A2)
        sage: f = H([x - y, x*y])
        sage: f([0, 1])
        (-1, 0)

    An example involving the projective line::

        sage: R.<x,y> = QQ[]
        sage: P1 = ProjectiveSpace(R)
        sage: H = P1.Hom(P1)
        sage: f = H([x^2 + y^2, x*y])
        sage: f([0, 1])
        (1 : 0)

    Some checks are performed to make sure the given polynomials
    define a morphism::

        sage: f = H([exp(x),exp(y)])                                                    # needs sage.symbolic
        Traceback (most recent call last):
        ...
        TypeError: polys (=[e^x, e^y]) must be elements of Multivariate
        Polynomial Ring in x, y over Rational Field
    """
    def __init__(self, parent, polys, check=True):
        """
        The Python constructor.

        See :class:`SchemeMorphism_polynomial` for details.

        EXAMPLES::

            sage: A2.<x,y> = AffineSpace(QQ, 2)
            sage: H = A2.Hom(A2)
            sage: H([x - y, x*y])
            Scheme endomorphism of Affine Space of dimension 2 over Rational Field
              Defn: Defined on coordinates by sending (x, y) to (x - y, x*y)
        """
        if check:
            if not isinstance(polys, (list, tuple)):
                raise TypeError("polys (=%s) must be a list or tuple" % polys)
            source_ring = parent.domain().ambient_space().coordinate_ring()
            target = parent._codomain.ambient_space()
            if len(polys) != target.ngens():
                raise ValueError("there must be %s polynomials" % target.ngens())
            F = []
            for poly in polys:
                try:
                    p = source_ring(poly)
                except TypeError:
                    try:
                        p = source_ring(poly.lift())
                    except (TypeError, AttributeError):
                        try:
                            p = source_ring(poly.numerator()) / source_ring(poly.denominator())
                        except (TypeError, AttributeError):
                            raise TypeError("polys (=%s) must be elements of %s" % (polys, source_ring))
                F.append(p)
            polys = Sequence(F)

        self._polys = tuple(polys)

        SchemeMorphism.__init__(self, parent)

    def __eq__(self, other):
        """
        Check equality of ``self`` and ``other``.

        INPUT:

        - ``other`` -- a morphism

        EXAMPLES::

            sage: A.<x,y> = AffineSpace(2, QQ)
            sage: I = A.identity_morphism()
            sage: I.parent().identity() == I
            True
        """
        if isinstance(other, SchemeMorphism_polynomial):
            if self.parent() == other.parent() and self._polys == other._polys:
                return True
        raise TypeError('cannot determine equality')

    def defining_polynomials(self):
        """
        Return the defining polynomials.

        OUTPUT:

        An immutable sequence of polynomials that defines this scheme
        morphism.

        EXAMPLES::

            sage: R.<x,y> = QQ[]
            sage: A.<x,y> = AffineSpace(R)
            sage: H = A.Hom(A)
            sage: H([x^3 + y, 1 - x - y]).defining_polynomials()
            (x^3 + y, -x - y + 1)
        """
        return self._polys

    def _call_(self, x):
        """
        Apply this morphism to a point in the domain.

        INPUT:

        - ``x`` -- a point in the domain or a list or tuple that
          defines a point in the domain

        OUTPUT: a point in the codomain

        EXAMPLES::

            sage: R.<x,y> = QQ[]
            sage: A.<x,y> = AffineSpace(R)
            sage: H = A.Hom(A)
            sage: f = H([y, x^2 + y])
            sage: f([2, 3])    # indirect doctest
            (3, 7)

        An example with algebraic schemes::

            sage: A.<x,y> = AffineSpace(QQ, 2)
            sage: X = A.subscheme(x)
            sage: Y = A.subscheme(y)
            sage: Hom_XY = X.Hom(Y)
            sage: f = Hom_XY([y,0])   # (0,y) |-> (y,0)
            sage: f
            Scheme morphism:
              From: Closed subscheme of Affine Space of dimension 2 over Rational Field
                    defined by: x
              To:   Closed subscheme of Affine Space of dimension 2 over Rational Field
                    defined by: y
              Defn: Defined on coordinates by sending (x, y) to (y, 0)
            sage: f([0,3])
            (3, 0)

        The input must be convertible into the map's domain::

            sage: f(0)
            Traceback (most recent call last):
            ...
            TypeError: 0 fails to convert into the map's domain Closed
            subscheme of Affine Space of dimension 2 over Rational Field
            defined by:
              x, but a `pushforward` method is not properly implemented

        It is possible to avoid the checks on the resulting point which can be
        useful for indeterminacies, but be careful!! ::

            sage: PS.<x,y> = ProjectiveSpace(QQ, 1)
            sage: H = Hom(PS, PS)
            sage: f = H([x^3, x*y^2])
            sage: P = PS(0,1)
            sage: f(P, check=False)
            (0 : 0)

            sage: P.<x,y,z> = ProjectiveSpace(ZZ, 2)
            sage: X = P.subscheme(x^2 - y^2)
            sage: H = Hom(X, X)
            sage: f = H([x^2, y^2, z^2])
            sage: f([4,4,1])
            (16 : 16 : 1)

            sage: P.<x,y,z> = ProjectiveSpace(ZZ, 2)
            sage: X = P.subscheme(x^2 - y^2)
            sage: H = Hom(X, X)
            sage: f = H([x^2, y^2, z^2])
            sage: f(P([4,4,1]))
            (16 : 16 : 1)
        """
        # Checks were done in __call__
        P = [f(x._coords) for f in self.defining_polynomials()]
        return self._codomain.point(P,check=True)

    def _call_with_args(self, x, args, kwds):
        """
        Apply this morphism to a point in the domain, with additional arguments.

        INPUT:

        - ``x`` -- a point in the domain or a list or tuple that defines a
          point in the domain
        - ``check`` -- boolean; either provided by position or name

        OUTPUT: a point in the codomain

        EXAMPLES::

            sage: R.<x,y> = QQ[]
            sage: A.<x,y> = AffineSpace(R)
            sage: H = A.Hom(A)
            sage: f = H([y, x^2 + y])
            sage: f([2,3])
            (3, 7)

        An example with algebraic schemes::

            sage: A.<x,y> = AffineSpace(QQ, 2)
            sage: X = A.subscheme(x)
            sage: Y = A.subscheme(y)
            sage: Hom_XY = X.Hom(Y)
            sage: f = Hom_XY([y,0])   # (0,y) |-> (y,0)
            sage: f
            Scheme morphism:
              From: Closed subscheme of Affine Space of dimension 2 over Rational Field
                    defined by: x
              To:   Closed subscheme of Affine Space of dimension 2 over Rational Field
                    defined by: y
              Defn: Defined on coordinates by sending (x, y) to (y, 0)
            sage: f([0,3])
            (3, 0)

        As usual, if the input does not belong to a map's domain, it is first
        attempted to convert it::

            sage: f(0)
            Traceback (most recent call last):
            ...
            TypeError: 0 fails to convert into the map's domain Closed subscheme of
            Affine Space of dimension 2 over Rational Field defined by:
              x, but a `pushforward` method is not properly implemented

        It is possible to avoid the checks on the resulting point which can be
        useful for indeterminacies, but be careful!! ::

            sage: PS.<x,y> = ProjectiveSpace(QQ, 1)
            sage: H = Hom(PS,PS)
            sage: f = H([x^3, x*y^2])
            sage: P = PS(0,1)
            sage: f(P, check=False)     # indirect doctest
            (0 : 0)

        ::

            sage: P.<x,y,z> = ProjectiveSpace(ZZ, 2)
            sage: X = P.subscheme(x^2 - y^2)
            sage: H = Hom(X, X)
            sage: f = H([x^2, y^2, z^2])
            sage: f([4,4,1])
            (16 : 16 : 1)

        ::

            sage: P.<x,y,z> = ProjectiveSpace(ZZ, 2)
            sage: P2.<u,v,w,t> = ProjectiveSpace(ZZ, 3)
            sage: X = P.subscheme(x^2 - y^2)
            sage: H = Hom(X, X)
            sage: f = H([x^2, y^2, z^2])
            sage: f(P2([4,4,1,1]))
            Traceback (most recent call last):
            ...
            TypeError: (4 : 4 : 1 : 1) fails to convert into the map's domain Closed subscheme of
            Projective Space of dimension 2 over Integer Ring defined by:
              x^2 - y^2, but a `pushforward` method is not properly implemented
        """
        if args:
            check = args[0]
        else:
            check = kwds.get("check", False)
        # containment of x in the domain has already been checked, in __call__
        P = [f(x._coords) for f in self.defining_polynomials()]
        return self._codomain.point(P,check)

    def _repr_defn(self):
        """
        Return a string representation of the definition of ``self``.

        OUTPUT: string

        EXAMPLES::

            sage: R.<x,y> = QQ[]
            sage: A.<x,y> = AffineSpace(R)
            sage: H = A.Hom(A)
            sage: f = H([y, x^2 + y])
            sage: print(f._repr_defn())
            Defined on coordinates by sending (x, y) to (y, x^2 + y)
        """
        i = self.domain().ambient_space()._repr_generic_point()
        o = self._codomain.ambient_space()._repr_generic_point(self.defining_polynomials())
        return "Defined on coordinates by sending %s to\n%s" % (i,o)

    def __getitem__(self, i):
        """
        Return the i-th poly with ``self[i]``.

        INPUT:

        - ``i`` -- integer

        OUTPUT: element of the coordinate ring of the domain

        EXAMPLES::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: H = Hom(P, P)
            sage: f = H([3/5*x^2, 6*y^2])
            sage: f[1]
            6*y^2
        """
        return self._polys[i]

    def __copy__(self):
        r"""
        Return a copy of ``self``.

        OUTPUT: :class:`SchemeMorphism_polynomial`

        EXAMPLES::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: H = Hom(P, P)
            sage: f = H([3/5*x^2, 6*y^2])
            sage: g = copy(f)
            sage: f == g
            True
            sage: f is g
            False

        ::

            sage: P.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: X = P.subscheme(x^2 - y^2)
            sage: Q = X(23, 23, 46)
            sage: P = X(1, 1, 1)
            sage: P != Q
            True
        """
        return self.parent()(self._polys)

    def coordinate_ring(self):
        r"""
        Return the coordinate ring of the ambient projective space.

        OUTPUT: a multivariable polynomial ring over the base ring

        EXAMPLES::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: H = Hom(P, P)
            sage: f = H([3/5*x^2, 6*y^2])
            sage: f.coordinate_ring()
            Multivariate Polynomial Ring in x, y over Rational Field

        ::

            sage: R.<t> = PolynomialRing(ZZ, 1)
            sage: P.<x,y> = ProjectiveSpace(R, 1)
            sage: H = Hom(P, P)
            sage: f = H([3*x^2, y^2])
            sage: f.coordinate_ring()
            Multivariate Polynomial Ring in x, y over Multivariate Polynomial Ring
            in t over Integer Ring
        """
        return self._polys[0].parent()

    def change_ring(self, R, check=True):
        r"""
        Return a new :class:`SchemeMorphism_polynomial` which is this map
        coerced to ``R``.

        If ``check`` is ``True``, then the initialization checks are performed.

        INPUT:

        - ``R`` -- ring or morphism

        - ``check`` -- boolean

        OUTPUT: a new :class:`SchemeMorphism_polynomial` which is this map
        coerced to ``R``

        TESTS::

            sage: # needs sage.rings.number_field
            sage: R.<t> = QQ[]
            sage: K.<v> = QuadraticField(2)
            sage: K2.<w> = NumberField(t**4 - 2)
            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: phi = K.embeddings(K2)[0]
            sage: f = DynamicalSystem_projective([x**2 + 3*y**2, y**2])                 # needs sage.schemes
            sage: f.change_ring(phi)                                                    # needs sage.schemes
            Dynamical System of Projective Space of dimension 1 over
             Number Field in w with defining polynomial t^4 - 2
              Defn: Defined on coordinates by sending (x : y) to (x^2 + 3*y^2 : y^2)

        ::

            sage: # needs sage.rings.number_field
            sage: R.<t> = QQ[]
            sage: K.<u> = QuadraticField(2)
            sage: K1.<v> = NumberField(t^4 - 2)
            sage: K2.<w> = NumberField(t^8 - 2)
            sage: P.<x,y> = ProjectiveSpace(K, 1)
            sage: phi = K1.embeddings(K2)[0]
            sage: f = DynamicalSystem_projective([x^2 + 3*y^2, y^2])                    # needs sage.schemes
            sage: f.change_ring(phi)                                                    # needs sage.schemes
            Traceback (most recent call last):
            ...
            ValueError: no canonical coercion of base ring of morphism to domain of embedding

        EXAMPLES::

            sage: P.<x,y> = ProjectiveSpace(ZZ, 1)
            sage: H = Hom(P, P)
            sage: f = H([3*x^2, y^2])
            sage: f.change_ring(GF(3))
            Scheme endomorphism of Projective Space of dimension 1 over Finite Field of size 3
              Defn: Defined on coordinates by sending (x : y) to (0 : y^2)

        ::

            sage: P.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: H = Hom(P, P)
            sage: f = H([5/2*x^3 + 3*x*y^2 - y^3, 3*z^3 + y*x^2, x^3 - z^3])
            sage: f.change_ring(GF(3))
            Scheme endomorphism of Projective Space of dimension 2 over Finite Field of size 3
              Defn: Defined on coordinates by sending (x : y : z) to
                    (x^3 - y^3 : x^2*y : x^3 - z^3)

        ::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: X = P.subscheme([5*x^2 - y^2])
            sage: H = Hom(X, X)
            sage: f = H([x, y])
            sage: f.change_ring(GF(3))
            Scheme endomorphism of Closed subscheme of Projective Space of dimension 1
            over Finite Field of size 3 defined by: -x^2 - y^2
              Defn: Defined on coordinates by sending (x : y) to (x : y)


        Check that :issue:`16834` is fixed::

            sage: # needs sage.rings.real_mpfr
            sage: A.<x,y,z> = AffineSpace(RR, 3)
            sage: h = Hom(A, A)
            sage: f = h([x^2 + 1.5, y^3, z^5 - 2.0])
            sage: f.change_ring(CC)
            Scheme endomorphism of Affine Space of dimension 3 over
             Complex Field with 53 bits of precision
              Defn: Defined on coordinates by sending (x, y, z) to
                    (x^2 + 1.50000000000000, y^3, z^5 - 2.00000000000000)

        ::

            sage: A.<x,y> = AffineSpace(ZZ, 2)
            sage: B.<u,v> = ProjectiveSpace(QQ, 1)
            sage: h = Hom(A,B)
            sage: f = h([x^2, y^2])
            sage: f.change_ring(QQ)
            Scheme morphism:
              From: Affine Space of dimension 2 over Rational Field
              To:   Projective Space of dimension 1 over Rational Field
              Defn: Defined on coordinates by sending (x, y) to (x^2 : y^2)

        ::

            sage: A.<x,y> = AffineSpace(QQ, 2)
            sage: H = Hom(A, A)
            sage: f = H([3*x^2/y, y^2/x])
            sage: f.change_ring(RR)                                                     # needs sage.rings.real_mpfr
            Scheme endomorphism of Affine Space of dimension 2 over Real Field with
            53 bits of precision
              Defn: Defined on coordinates by sending (x, y) to
                    (3.00000000000000*x^2/y, y^2/x)

        ::

            sage: # needs sage.rings.number_field
            sage: R.<x> = PolynomialRing(QQ)
            sage: K.<a> = NumberField(x^3 - x + 1)
            sage: P.<x,y> = ProjectiveSpace(K, 1)
            sage: H = End(P)
            sage: f = H([x^2 + a*x*y + a^2*y^2, y^2])
            sage: emb = K.embeddings(QQbar)
            sage: f.change_ring(emb[0])
            Scheme endomorphism of Projective Space of dimension 1
             over Algebraic Field
               Defn: Defined on coordinates by sending (x : y) to
                     (x^2 + (-1.324717957244746?)*x*y + 1.754877666246693?*y^2 : y^2)
            sage: f.change_ring(emb[1])
            Scheme endomorphism of Projective Space of dimension 1
             over Algebraic Field
               Defn: Defined on coordinates by sending (x : y) to
                     (x^2 + (0.6623589786223730? - 0.5622795120623013?*I)*x*y
                      + (0.1225611668766537? - 0.744861766619745?*I)*y^2 : y^2)

        ::

            sage: # needs sage.rings.number_field sage.symbolic
            sage: K.<v> = QuadraticField(2, embedding=QQbar(sqrt(2)))
            sage: P.<x,y> = ProjectiveSpace(K, 1)
            sage: H = End(P)
            sage: f = H([x^2 + v*y^2, y^2])
            sage: f.change_ring(QQbar)
            Scheme endomorphism of Projective Space of dimension 1
             over Algebraic Field
              Defn: Defined on coordinates by sending (x : y) to
                    (x^2 + 1.414213562373095?*y^2 : y^2)

        ::

            sage: # needs sage.rings.number_field sage.symbolic
            sage: from sage.misc.verbose import set_verbose
            sage: set_verbose(-1)
            sage: K.<w> = QuadraticField(2, embedding=QQbar(-sqrt(2)))
            sage: P.<x,y> = ProjectiveSpace(K, 1)
            sage: X = P.subscheme(x - y)
            sage: H = End(X)
            sage: f = H([6*x^2 + 2*x*y + 16*y^2, -w*x^2 - 4*x*y - 4*y^2])
            sage: f.change_ring(QQbar)
            Scheme endomorphism of Closed subscheme of Projective Space of dimension 1
             over Algebraic Field defined by: x - y
              Defn: Defined on coordinates by sending (x : y) to
                    (6*x^2 + 2*x*y + 16*y^2 : 1.414213562373095?*x^2 + (-4)*x*y + (-4)*y^2)

        ::

            sage: # needs sage.rings.number_field
            sage: R.<x> = QQ[]
            sage: f = x^6 - 2
            sage: L.<b> = NumberField(f, embedding=f.roots(QQbar)[1][0])
            sage: A.<x,y> = AffineSpace(L, 2)
            sage: H = Hom(A, A)
            sage: F = H([b*x/y, 1 + y])
            sage: F.change_ring(QQbar)
            Scheme endomorphism of Affine Space of dimension 2 over Algebraic Field
              Defn: Defined on coordinates by sending (x, y) to
                    (1.122462048309373?*x/y, y + 1)

        ::

            sage: # needs sage.rings.number_field
            sage: K.<a> = QuadraticField(-1)
            sage: A.<x,y> = AffineSpace(K, 2)
            sage: H = End(A)
            sage: phi = H([x/y, y])
            sage: emb = K.embeddings(QQbar)[0]
            sage: phi.change_ring(emb)
            Scheme endomorphism of Affine Space of dimension 2 over Algebraic Field
              Defn: Defined on coordinates by sending (x, y) to (x/y, y)
        """
        T = self.domain().change_ring(R)
        if self.is_endomorphism():
            H = End(T)
        else:
            S = self.codomain().change_ring(R)
            H = Hom(T, S)

        if isinstance(R, Map):
            from sage.structure.coerce_maps import CallableConvertMap
            if R.domain() == self.base_ring():
                S = self.domain().ambient_space().coordinate_ring()
                T = T.ambient_space().coordinate_ring()
                phi = CallableConvertMap(S, T, lambda self, g:T(g.map_coefficients(R)))
                G = []
                for f in self:
                    if isinstance(f, FractionFieldElement):
                        G.append(phi(f.numerator())/phi(f.denominator()))
                    else:
                        G.append(phi(f))
            elif R.domain().coerce_map_from(self.base_ring()) is not None:
                R = R*R.domain().coerce_map_from(self.base_ring())
                S = self.domain().ambient_space().coordinate_ring()
                T = T.ambient_space().coordinate_ring()
                phi = CallableConvertMap(S, T, lambda self, g:T(g.map_coefficients(R)))
                G = []
                for f in self:
                    if isinstance(f, FractionFieldElement):
                        G.append(phi(f.numerator())/phi(f.denominator()))
                    else:
                        G.append(phi(f))
            else:
                raise ValueError("no canonical coercion of base ring of morphism to domain of embedding")

        else:
            G = []
            for f in self:
                if isinstance(f, FractionFieldElement):
                    G.append(f.numerator().change_ring(R) / f.denominator().change_ring(R))
                else:
                    G.append(f.change_ring(R))
        return H(G, check)

    def specialization(self, D=None, phi=None, homset=None):
        r"""
        Specialization of this map.

        Given a family of maps defined over a polynomial ring. A specialization
        is a particular member of that family. The specialization can be specified either
        by a dictionary or a :class:`SpecializationMorphism`.

        INPUT:

        - ``D`` -- dictionary (optional)

        - ``phi`` -- SpecializationMorphism (optional)

        - ``homset`` -- homset of specialized map (optional)

        OUTPUT: :class:`SchemeMorphism_polynomial`

        EXAMPLES::

            sage: R.<c> = PolynomialRing(QQ)
            sage: P.<x,y> = ProjectiveSpace(R, 1)
            sage: H = End(P)
            sage: f = H([x^2 + c*y^2, y^2])
            sage: f.specialization({c: 1})
            Scheme endomorphism of Projective Space of dimension 1 over Rational Field
              Defn: Defined on coordinates by sending (x : y) to (x^2 + y^2 : y^2)

        ::

            sage: R.<a,b> = PolynomialRing(QQ)
            sage: P.<x,y> = ProjectiveSpace(R, 1)
            sage: H = End(P)
            sage: f = H([x^3 + a*x*y^2 + b*y^3, y^3])
            sage: from sage.rings.polynomial.flatten import SpecializationMorphism
            sage: phi = SpecializationMorphism(P.coordinate_ring(), {a: 2, b: -1})
            sage: F = f.specialization(phi=phi); F
            Scheme endomorphism of Projective Space of dimension 1 over Rational Field
              Defn: Defined on coordinates by sending (x : y) to
                    (x^3 + 2*x*y^2 - y^3 : y^3)
            sage: g = H([x^2 + a*y^2, y^2])
            sage: G = g.specialization(phi=phi)
            sage: G.parent() is F.parent()
            True
            sage: G = g.specialization(phi=phi, homset=F.parent())
            sage: G.parent() is F.parent()
            True

        ::

            sage: R.<c> = PolynomialRing(QQ)
            sage: P.<x,y> = ProjectiveSpace(R, 1)
            sage: X = P.subscheme([x - c*y])
            sage: H = End(X)
            sage: f = H([x^2, c*y^2])
            sage: f.specialization({c: 2})
            Scheme endomorphism of Closed subscheme of Projective Space of dimension 1
             over Rational Field defined by: x - 2*y
                  Defn: Defined on coordinates by sending (x : y) to (x^2 : 2*y^2)

        ::

            sage: # needs sage.schemes
            sage: R.<c> = QQ[]
            sage: P.<x,y> = ProjectiveSpace(R, 1)
            sage: f = DynamicalSystem_projective([x^2 + c*y^2, y^2], domain=P)
            sage: F = f.dynatomic_polynomial(3)                                         # needs sage.libs.pari
            sage: g = F.specialization({c: 1}); g
            x^6 + x^5*y + 4*x^4*y^2 + 3*x^3*y^3 + 7*x^2*y^4 + 4*x*y^5 + 5*y^6
            sage: g == f.specialization({c:1}).dynatomic_polynomial(3)                  # needs sage.libs.pari
            True

        ::

            sage: R1.<alpha, beta> = QQ[]
            sage: A.<x> = AffineSpace(Frac(R1), 1)
            sage: f = DynamicalSystem_affine([alpha/(x^2 + 1/alpha)/(x - 1/beta^2)])
            sage: f.specialization({alpha: 5, beta: 10})
            Dynamical System of Affine Space of dimension 1 over Rational Field
              Defn: Defined on coordinates by sending (x) to
                      (5/(x^3 - 1/100*x^2 + 1/5*x - 1/500))
            sage: f_5_10 = f.specialization({alpha: 5}).specialization({beta: 10})
            sage: f_5_10 == f.specialization({alpha: 5, beta: 10})
            True
        """
        if D is None:
            if phi is None:
                raise ValueError("either the dictionary or the specialization must be provided")
        else:
            if isinstance(self[0].parent(), FractionField_generic):
                from sage.rings.polynomial.flatten import FractionSpecializationMorphism
                phi = FractionSpecializationMorphism(self[0].parent(), D)
            else:
                from sage.rings.polynomial.flatten import SpecializationMorphism
                phi = SpecializationMorphism(self[0].parent(), D)
        if homset is None:
            domain = self.domain()
            if isinstance(domain, AlgebraicScheme_subscheme):
                domain = domain.specialization(phi=phi)
            else:
                domain = domain.change_ring(phi.codomain().base_ring())
            if self.is_endomorphism():
                homset = End(domain)
            else:
                codomain = self.codomain()
                if isinstance(codomain, AlgebraicScheme_subscheme):
                    codomain = codomain.specialization(phi=phi)
                else:
                    codomain = codomain.change_ring(phi.codomain().base_ring())
                homset = Hom(domain, codomain)
        return homset([phi(g) for g in self])

    def _composition_(self, other, homset):
        r"""
        Straightforward implementation of composition for scheme morphisms
        defined by polynomials.

        TESTS::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: H = Hom(P, P)
            sage: f = H([x^2 - 29/16*y^2, y^2])
            sage: g = H([y, x + y])
            sage: h = f*g
            sage: h
            Scheme endomorphism of Projective Space of dimension 1 over Rational Field
              Defn: Defined on coordinates by sending (x : y) to
                    (-29/16*x^2 - 29/8*x*y - 13/16*y^2 : x^2 + 2*x*y + y^2)
            sage: p = P((1,3))
            sage: h(p) == f(g(p))
            True

            sage: Q = ProjectiveSpace(QQ, 2)
            sage: H2 = Hom(P, Q)
            sage: h2 = H2([x^2 + y^2, x^2, y^2 + 2*x^2])
            sage: h2 * f
            Scheme morphism:
              From: Projective Space of dimension 1 over Rational Field
              To:   Projective Space of dimension 2 over Rational Field
              Defn: Defined on coordinates by sending (x : y) to
                    (x^4 - 29/8*x^2*y^2 + 1097/256*y^4
                     : x^4 - 29/8*x^2*y^2 + 841/256*y^4
                     : 2*x^4 - 29/4*x^2*y^2 + 969/128*y^4)

        ::

            sage: A.<x,y> = AffineSpace(QQ, 2)
            sage: A1.<z> = AffineSpace(QQ, 1)
            sage: H = End(A)
            sage: f = H([x^2 + y^2, y^2/x])
            sage: H1 = Hom(A, A1)
            sage: g = H1([x + y^2])
            sage: g*f
            Scheme morphism:
              From: Affine Space of dimension 2 over Rational Field
              To:   Affine Space of dimension 1 over Rational Field
              Defn: Defined on coordinates by sending (x, y) to
                    ((x^4 + x^2*y^2 + y^4)/x^2)
            sage: f*g
            Traceback (most recent call last):
            ...
            TypeError: self (=Scheme endomorphism of Affine Space of dimension 2 over Rational Field
              Defn: Defined on coordinates by sending (x, y) to
                    (x^2 + y^2, y^2/x)) domain
            must equal right (=Scheme morphism:
              From: Affine Space of dimension 2 over Rational Field
              To:   Affine Space of dimension 1 over Rational Field
              Defn: Defined on coordinates by sending (x, y) to
                    (y^2 + x)) codomain

        Not both defined by polynomials::

            sage: # needs sage.rings.number_field
            sage: x = polygen(QQ)
            sage: K.<a> = NumberField(x^2 - 2)
            sage: p1, p2 = K.Hom(K)
            sage: R.<x,y> = K[]
            sage: q1 = R.Hom(R)(p1)
            sage: A = AffineSpace(R)
            sage: f1 = A.Hom(A)(q1)
            sage: g = A.Hom(A)([x^2 - y, y + 1])
            sage: g*f1
            Composite map:
              From: Affine Space of dimension 2 over Number Field in a
                    with defining polynomial x^2 - 2
              To:   Affine Space of dimension 2 over Number Field in a
                    with defining polynomial x^2 - 2
              Defn:   Generic endomorphism of Affine Space of dimension 2
                       over Number Field in a with defining polynomial x^2 - 2
                    then
                      Generic endomorphism of Affine Space of dimension 2
                       over Number Field in a with defining polynomial x^2 - 2
        """
        try:
            opolys = tuple(other._polys)
        except AttributeError:
            return super()._composition_(other, homset)
        return homset([p(*opolys) for p in self._polys])


class SchemeMorphism_polynomial_id(SchemeMorphism_id, SchemeMorphism_polynomial):
    """
    Return the identity morphism from `X` to itself.

    INPUT:

    - ``X`` -- an affine or projective scheme

    EXAMPLES::

        sage: X = Spec(ZZ)
        sage: X.identity_morphism()  # indirect doctest
        Scheme endomorphism of Spectrum of Integer Ring
          Defn: Identity map
    """
    def __init__(self, X):
        """
        Initialize.

        TESTS::

            sage: A = AffineSpace(2, GF(3))
            sage: A.identity_morphism().defining_polynomials()
            (x0, x1)
        """
        super().__init__(X)
        variables = X.ambient_space().coordinate_ring().gens()
        SchemeMorphism_polynomial.__init__(self, X.Hom(X), variables, check=False)


############################################################################
# Rational points on schemes, which we view as morphisms determined
# by coordinates.
############################################################################

class SchemeMorphism_point(SchemeMorphism):
    r"""
    Base class for rational points on schemes.

    Recall that the `K`-rational points of a scheme `X` over `k` can
    be identified with the set of morphisms `\mathrm{Spec}(K) \to X`. In Sage,
    the rational points are implemented by such scheme morphisms.

    EXAMPLES::

        sage: from sage.schemes.generic.morphism import SchemeMorphism
        sage: f = SchemeMorphism(Spec(ZZ).Hom(Spec(ZZ)))
        sage: type(f)
        <class 'sage.schemes.generic.morphism.SchemeMorphism'>
    """
    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        OUTPUT: string

        EXAMPLES::

            sage: A = AffineSpace(2, QQ)
            sage: a = A(1,2)
            sage: a._repr_()
            '(1, 2)'
        """
        return self._codomain.ambient_space()._repr_generic_point(self._coords)

    def _latex_(self):
        r"""
        Return a latex representation of ``self``.

        OUTPUT: string

        EXAMPLES::

            sage: A = AffineSpace(2, QQ)
            sage: a = A(1,2)
            sage: latex(a) == a._latex_()
            True
            sage: a._latex_()
            '\\left(1, 2\\right)'
        """
        return self._codomain.ambient_space()._latex_generic_point(self._coords)

    def __getitem__(self, n):
        """
        Return the ``n``-th coordinate.

        OUTPUT: the coordinate values as an element of the base ring

        EXAMPLES::

            sage: A = AffineSpace(2, QQ)
            sage: a = A(1,2)
            sage: a[0]
            1
            sage: a[1]
            2
        """
        return self._coords[n]

    def __iter__(self):
        """
        Iterate over the coordinates of the point.

        OUTPUT: an iterator

        EXAMPLES::

            sage: A = AffineSpace(2, QQ)
            sage: a = A(1,2)
            sage: iter = a.__iter__()
            sage: next(iter)
            1
            sage: next(iter)
            2
            sage: list(a)
            [1, 2]
        """
        return iter(self._coords)

    def __tuple__(self):
        """
        Return the coordinates as a tuple.

        OUTPUT: a tuple

        EXAMPLES::

            sage: A = AffineSpace(2, QQ)
            sage: a = A(1,2)
            sage: tuple(a)
            (1, 2)
        """
        return self._coords

    def __len__(self):
        """
        Return the number of coordinates.

        OUTPUT: integer. The number of coordinates used to describe the point

        EXAMPLES::

            sage: A = AffineSpace(2, QQ)
            sage: a = A(1,2)
            sage: len(a)
            2
        """
        return len(self._coords)

    def _richcmp_(self, other, op):
        """
        Compare two scheme morphisms.

        INPUT:

        - ``other`` -- anything; to compare against the scheme
          morphism ``self``

        OUTPUT: boolean

        EXAMPLES::

            sage: A = AffineSpace(2, QQ)
            sage: a = A(1,2)
            sage: b = A(3,4)
            sage: a < b
            True
            sage: a != b
            True
        """
        if not isinstance(other, SchemeMorphism_point):
            try:
                other = self._codomain.ambient_space()(other)
            except TypeError:
                return NotImplemented
        return richcmp(self._coords, other._coords, op)

    def scheme(self):
        """
        Return the scheme whose point is represented.

        OUTPUT: a scheme

        EXAMPLES::

            sage: A = AffineSpace(2, QQ)
            sage: a = A(1,2)
            sage: a.scheme()
            Affine Space of dimension 2 over Rational Field
        """
        return self._codomain

    def change_ring(self, R, check=True):
        r"""
        Return a new :class:`SchemeMorphism_point` which is this point coerced
        to ``R``.

        If ``check`` is true, then the initialization checks are performed.

        INPUT:

        - ``R`` -- ring or morphism

        - ``check`` -- boolean

        OUTPUT: :class:`SchemeMorphism_point`

        EXAMPLES::

            sage: P.<x,y,z> = ProjectiveSpace(ZZ, 2)
            sage: X = P.subscheme(x^2 - y^2)
            sage: X(23,23,1).change_ring(GF(13))
            (10 : 10 : 1)

        ::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: P(-2/3,1).change_ring(CC)                                             # needs sage.rings.real_mpfr
            (-0.666666666666667 : 1.00000000000000)

        ::

            sage: P.<x,y> = ProjectiveSpace(ZZ, 1)
            sage: P(152,113).change_ring(Zp(5))                                         # needs sage.rings.padics
            (2 + 5^2 + 5^3 + O(5^20) : 3 + 2*5 + 4*5^2 + O(5^20))

        ::

            sage: # needs sage.rings.number_field
            sage: K.<v> = QuadraticField(-7)
            sage: O = K.maximal_order()
            sage: P.<x,y> = ProjectiveSpace(O, 1)
            sage: H = End(P)
            sage: F = H([x^2 + O(v)*y^2, y^2])
            sage: F.change_ring(K).change_ring(K.embeddings(QQbar)[0])
            Scheme endomorphism of Projective Space of dimension 1
             over Algebraic Field
              Defn: Defined on coordinates by sending (x : y) to
                    (x^2 + (-2.645751311064591?*I)*y^2 : y^2)

        ::

            sage: # needs sage.rings.number_field
            sage: R.<x> = PolynomialRing(QQ)
            sage: K.<a> = NumberField(x^2 - x + 1)
            sage: P.<x,y> = ProjectiveSpace(K, 1)
            sage: Q = P([a + 1, 1])
            sage: emb = K.embeddings(QQbar)
            sage: Q.change_ring(emb[0])
            (1.5000000000000000? - 0.866025403784439?*I : 1)
            sage: Q.change_ring(emb[1])
            (1.5000000000000000? + 0.866025403784439?*I : 1)

        ::

            sage: # needs sage.rings.number_field
            sage: K.<v> = QuadraticField(2)
            sage: P.<x,y> = ProjectiveSpace(K, 1)
            sage: Q = P([v,1])
            sage: Q.change_ring(K.embeddings(QQbar)[0])
            (-1.414213562373095? : 1)

        ::

            sage: # needs sage.rings.number_field
            sage: R.<x> = QQ[]
            sage: f = x^6 - 2
            sage: L.<b> = NumberField(f, embedding=f.roots(QQbar)[1][0])
            sage: A.<x,y> = AffineSpace(L, 2)
            sage: P = A([b,1])
            sage: P.change_ring(QQbar)
            (1.122462048309373?, 1)
        """
        S = self.codomain().change_ring(R)
        Q = [R(t) for t in self]
        return S.point(Q, check=check)

    def __copy__(self):
        r"""
        Return a copy of the :class:`SchemeMorphism_point` ``self`` coerced to
        `R`.

        OUTPUT: :class:`SchemeMorphism_point`

        EXAMPLES::

            sage: P.<x,y> = ProjectiveSpace(ZZ, 1)
            sage: Q = P(152, 113)
            sage: Q2 = copy(Q)
            sage: Q2 is Q
            False
            sage: Q2 == Q
            True
        """
        return self._codomain.point(self._coords, check=False)

    def specialization(self, D=None, phi=None, ambient=None):
        r"""
        Specialization of this point.

        Given a family of points defined over a polynomial ring. A specialization
        is a particular member of that family. The specialization can be specified either
        by a dictionary or a :class:`SpecializationMorphism`.

        INPUT:

        - ``D`` -- dictionary (optional)

        - ``phi`` -- SpecializationMorphism (optional)

        - ``ambient`` -- ambient space of specialized point (optional)

        OUTPUT: :class:`SchemeMorphism_polynomial`

        EXAMPLES::

            sage: R.<c> = PolynomialRing(QQ)
            sage: P.<x,y> = ProjectiveSpace(R, 1)
            sage: Q = P([c,1])
            sage: Q.specialization({c: 1})
            (1 : 1)

        ::

            sage: R.<a,b> = PolynomialRing(QQ)
            sage: P.<x,y> = ProjectiveSpace(R, 1)
            sage: Q = P([a^2 + 2*a*b + 34, 1])
            sage: from sage.rings.polynomial.flatten import SpecializationMorphism
            sage: phi = SpecializationMorphism(P.coordinate_ring(), {a: 2, b: -1})
            sage: T = Q.specialization(phi=phi); T
            (34 : 1)
            sage: Q2 = P([a,1])
            sage: T2 = Q2.specialization(phi=phi)
            sage: T2.codomain() is T.codomain()
            True
            sage: T3 = Q2.specialization(phi=phi, ambient=T.codomain())
            sage: T3.codomain() is T.codomain()
            True

        ::

            sage: R.<c> = PolynomialRing(QQ)
            sage: P.<x,y> = ProjectiveSpace(R, 1)
            sage: X = P.subscheme([x - c*y])
            sage: Q = X([c, 1])
            sage: Q2 = Q.specialization({c:2}); Q2
            (2 : 1)
            sage: Q2.codomain()
            Closed subscheme of Projective Space of dimension 1 over Rational Field
             defined by: x - 2*y

        ::

            sage: R.<l> = PolynomialRing(QQ)
            sage: S.<k,j> = PolynomialRing(R)
            sage: K.<a,b,c,d> = S[]
            sage: P.<x,y> = ProjectiveSpace(K, 1)
            sage: H = End(P)
            sage: Q = P([a^2, b^2])
            sage: Q.specialization({a: 2})
            (4 : b^2)
        """
        if D is None:
            if phi is None:
                raise ValueError("either the dictionary or the specialization must be provided")
        else:
            from sage.rings.polynomial.flatten import SpecializationMorphism
            phi = SpecializationMorphism(self.codomain().ambient_space().coordinate_ring(), D)
        if ambient is None:
            ambient = self.codomain()
            if isinstance(ambient, AlgebraicScheme_subscheme):
                ambient = ambient.specialization(phi=phi)
            else:
                ambient = ambient.change_ring(phi.codomain().base_ring())
        psi = ambient.ambient_space().coordinate_ring().hom([0 for i in range(ambient.ambient_space().ngens())], ambient.base_ring())
        return ambient([psi(phi(t)) for t in self])



# sage_setup: distribution = sagemath-categories
from sage.rings.ring cimport Field

cdef class NumberField(Field):
    cdef int _embedded_real
    cdef list _gen_approx

    cpdef _get_embedding_approx(self, size_t i)



# sage_setup: distribution = sagemath-categories
"""
Space of homomorphisms between two rings
"""

#*****************************************************************************
#       Copyright (C) 2006 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.categories.homset import HomsetWithBase
from sage.categories.rings import Rings

_Rings = Rings()

from sage.rings import morphism, quotient_ring


def is_RingHomset(H):
    """
    Return ``True`` if ``H`` is a space of homomorphisms between two rings.

    EXAMPLES::

        sage: from sage.rings.homset import is_RingHomset as is_RH
        sage: is_RH(Hom(ZZ, QQ))
        doctest:warning...
        DeprecationWarning: the function is_RingHomset is deprecated;
        use 'isinstance(..., RingHomset_generic)' instead
        See https://github.com/sagemath/sage/issues/37922 for details.
        True
        sage: is_RH(ZZ)
        False
        sage: is_RH(Hom(RR, CC))                                                        # needs sage.rings.real_mpfr
        True
        sage: is_RH(Hom(FreeModule(ZZ,1), FreeModule(QQ,1)))                            # needs sage.modules
        False
    """
    from sage.misc.superseded import deprecation
    deprecation(37922, "the function is_RingHomset is deprecated; use 'isinstance(..., RingHomset_generic)' instead")
    return isinstance(H, RingHomset_generic)


def RingHomset(R, S, category=None):
    """
    Construct a space of homomorphisms between the rings ``R`` and ``S``.

    For more on homsets, see :func:`Hom()`.

    EXAMPLES::

        sage: Hom(ZZ, QQ) # indirect doctest
        Set of Homomorphisms from Integer Ring to Rational Field
    """
    if isinstance(R, quotient_ring.QuotientRing_nc):
        from .polynomial.polynomial_quotient_ring import PolynomialQuotientRing_generic
        if not isinstance(R, PolynomialQuotientRing_generic):  # backwards compatibility
            return RingHomset_quo_ring(R, S, category=category)
    return RingHomset_generic(R, S, category=category)


class RingHomset_generic(HomsetWithBase):
    """
    A generic space of homomorphisms between two rings.

    EXAMPLES::

        sage: Hom(ZZ, QQ)
        Set of Homomorphisms from Integer Ring to Rational Field
        sage: QQ.Hom(ZZ)
        Set of Homomorphisms from Rational Field to Integer Ring
    """

    Element = morphism.RingHomomorphism

    def __init__(self, R, S, category=None):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: Hom(ZZ, QQ)
            Set of Homomorphisms from Integer Ring to Rational Field
        """
        if category is None:
            category = _Rings
        HomsetWithBase.__init__(self, R, S, category)

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: Hom(ZZ, QQ) # indirect doctest
            Set of Homomorphisms from Integer Ring to Rational Field
        """
        return "Set of Homomorphisms from %s to %s" % (self.domain(), self.codomain())

    def has_coerce_map_from(self, x):
        """
        The default for coercion maps between ring homomorphism spaces is
        very restrictive (until more implementation work is done).

        Currently this checks if the domains and the codomains are equal.

        EXAMPLES::

            sage: H = Hom(ZZ, QQ)
            sage: H2 = Hom(QQ, ZZ)
            sage: H.has_coerce_map_from(H2)
            False
        """
        return (x.domain() == self.domain() and x.codomain() == self.codomain())

    def _element_constructor_(self, x, check=True, base_map=None):
        """
        Construct an element of ``self`` from ``x``.

        EXAMPLES::

            sage: H = Hom(ZZ, QQ)
            sage: phi = H([1]); phi
            Ring morphism:
              From: Integer Ring
              To:   Rational Field
              Defn: 1 |--> 1
            sage: H2 = Hom(QQ, QQ)
            sage: phi2 = H2(phi); phi2
            Ring endomorphism of Rational Field
              Defn: 1 |--> 1
            sage: H(phi2)
            Ring morphism:
              From: Integer Ring
              To:   Rational Field
              Defn: 1 |--> 1

        You can provide a morphism on the base::

            sage: # needs sage.rings.finite_rings
            sage: k = GF(9)
            sage: z2 = k.gen()
            sage: cc = k.frobenius_endomorphism()
            sage: R.<x> = k[]
            sage: H = Hom(R, R)
            sage: phi = H([x^2], base_map=cc); phi
            Ring endomorphism of Univariate Polynomial Ring in x
             over Finite Field in z2 of size 3^2
              Defn: x |--> x^2
                    with map of base ring
            sage: phi(z2 * x) == z2^3 * x^2
            True

            sage: R.<x> = ZZ[]
            sage: K.<a> = GF(7^2)                                                       # needs sage.rings.finite_rings
            sage: L.<u> = K.extension(x^3 - 3)                                          # needs sage.rings.finite_rings
            sage: phi = L.hom([u^7], base_map=K.frobenius_endomorphism())               # needs sage.rings.finite_rings
            sage: phi(u) == u^7                                                         # needs sage.rings.finite_rings
            True
            sage: phi(a) == a^7                                                         # needs sage.rings.finite_rings
            True

        TESTS::

            sage: H = Hom(ZZ, QQ)
            sage: H == loads(dumps(H))
            True
        """
        from sage.categories.map import Map
        # Case 0: the homomorphism is given by images of generators
        if not (isinstance(x, Map) and x.category_for().is_subcategory(Rings())):
            return morphism.RingHomomorphism_im_gens(self, x, base_map=base_map, check=check)
        if base_map is not None:
            raise ValueError("cannot specify base_map when providing a map")
        # Case 1: the parent fits
        if x.parent() == self:
            if isinstance(x, morphism.RingHomomorphism_im_gens):
                return morphism.RingHomomorphism_im_gens(self, x.im_gens())
            elif isinstance(x, morphism.RingHomomorphism_cover):
                return morphism.RingHomomorphism_cover(self)
            elif isinstance(x, morphism.RingHomomorphism_from_base):
                return morphism.RingHomomorphism_from_base(self, x.underlying_map())
        # Case 2: unique extension via fraction field
        try:
            if (isinstance(x, morphism.RingHomomorphism_im_gens)
                and x.domain().fraction_field().has_coerce_map_from(self.domain())):
                return morphism.RingHomomorphism_im_gens(self, x.im_gens())
        except (TypeError, ValueError):
            pass
        # Case 3: the homomorphism can be extended by coercion
        try:
            return x.extend_codomain(self.codomain()).extend_domain(self.domain())
        except (TypeError, ValueError):
            pass
        # Case 4: the homomorphism is induced from the base ring
        if (self.domain() != self.domain().base()
            or self.codomain() != self.codomain().base()):
            x = self.domain().base().Hom(self.codomain().base())(x)
            return morphism.RingHomomorphism_from_base(self, x)
        raise ValueError('cannot convert {} to an element of {}'.format(x, self))

    def natural_map(self):
        """
        Return the natural map from the domain to the codomain.

        The natural map is the coercion map from the domain ring to the
        codomain ring.

        EXAMPLES::

            sage: H = Hom(ZZ, QQ)
            sage: H.natural_map()
            Natural morphism:
              From: Integer Ring
              To:   Rational Field
        """
        f = self.codomain().coerce_map_from(self.domain())
        if f is None:
            raise TypeError("natural coercion morphism from %s to %s not defined" % (self.domain(), self.codomain()))
        return f

    def zero(self):
        r"""
        Return the zero element of this homset.

        EXAMPLES:

        Since a ring homomorphism maps 1 to 1, there can only be a zero
        morphism when mapping to the trivial ring::

            sage: Hom(ZZ, Zmod(1)).zero()
            Ring morphism:
              From: Integer Ring
              To:   Ring of integers modulo 1
              Defn: 1 |--> 0
            sage: Hom(ZZ, Zmod(2)).zero()
            Traceback (most recent call last):
            ...
            ValueError: homset has no zero element
        """
        if not self.codomain().is_zero():
            raise ValueError("homset has no zero element")
        # there is only one map in this homset
        return self.an_element()


class RingHomset_quo_ring(RingHomset_generic):
    """
    Space of ring homomorphisms where the domain is a (formal) quotient
    ring.

    EXAMPLES::

        sage: R.<x,y> = PolynomialRing(QQ, 2)
        sage: S.<a,b> = R.quotient(x^2 + y^2)                                           # needs sage.libs.singular
        sage: phi = S.hom([b,a]); phi                                                   # needs sage.libs.singular
        Ring endomorphism of Quotient of Multivariate Polynomial Ring in x, y
         over Rational Field by the ideal (x^2 + y^2)
          Defn: a |--> b
                b |--> a
        sage: phi(a)                                                                    # needs sage.libs.singular
        b
        sage: phi(b)                                                                    # needs sage.libs.singular
        a

    TESTS:

    We test pickling of a homset from a quotient.

    ::

        sage: # needs sage.libs.singular
        sage: R.<x,y> = PolynomialRing(QQ, 2)
        sage: S.<a,b> = R.quotient(x^2 + y^2)
        sage: H = S.Hom(R)
        sage: H == loads(dumps(H))
        True

    We test pickling of actual homomorphisms in a quotient::

        sage: phi = S.hom([b,a])                                                        # needs sage.libs.singular
        sage: phi == loads(dumps(phi))                                                  # needs sage.libs.singular
        True
    """

    Element = morphism.RingHomomorphism_from_quotient

    def _element_constructor_(self, x, base_map=None, check=True):
        """
        Construct an element of ``self`` from ``x``.

        EXAMPLES::

            sage: # needs sage.libs.singular
            sage: R.<x,y> = PolynomialRing(QQ, 2)
            sage: S.<a,b> = R.quotient(x^2 + y^2)
            sage: H = S.Hom(R)
            sage: phi = H([b, a]); phi
            Ring morphism:
              From: Quotient of Multivariate Polynomial Ring in x, y over Rational Field by the ideal (x^2 + y^2)
              To:   Multivariate Polynomial Ring in x, y over Rational Field
              Defn: a |--> b
                    b |--> a
            sage: R2.<x,y> = PolynomialRing(ZZ, 2)
            sage: H2 = Hom(R2, S)
            sage: H2(phi)
            Composite map:
              From: Multivariate Polynomial Ring in x, y over Integer Ring
              To:   Quotient of Multivariate Polynomial Ring in x, y over Rational Field by the ideal (x^2 + y^2)
              Defn:   Coercion map:
                      From: Multivariate Polynomial Ring in x, y over Integer Ring
                      To:   Quotient of Multivariate Polynomial Ring in x, y over Rational Field by the ideal (x^2 + y^2)
                    then
                      Ring morphism:
                      From: Quotient of Multivariate Polynomial Ring in x, y over Rational Field by the ideal (x^2 + y^2)
                      To:   Multivariate Polynomial Ring in x, y over Rational Field
                      Defn: a |--> b
                            b |--> a
                    then
                      Coercion map:
                      From: Multivariate Polynomial Ring in x, y over Rational Field
                      To:   Quotient of Multivariate Polynomial Ring in x, y over Rational Field by the ideal (x^2 + y^2)
        """
        if isinstance(x, morphism.RingHomomorphism_from_quotient):
            phi = x._phi()
        else:
            pi = self.domain().cover()
            phi = pi.domain().hom(x, base_map=base_map, check=check)
        return self.element_class(self, phi)



# sage_setup: distribution = sagemath-categories
# sage.doctest: optional - ptyprocess
"""
Sage wrapper around pexpect's ``spawn`` class and
the ptyprocess's ``PtyProcess`` class.

AUTHOR:

- Jeroen Demeyer (2015-02-01): initial version, see :issue:`17686`.

- Jeroen Demeyer (2015-12-04): add support for pexpect 4 + ptyprocess,
  see :issue:`10295`.
"""

# ***************************************************************************
#       Copyright (C) 2015 Jeroen Demeyer <jdemeyer@cage.ugent.be>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ***************************************************************************

from pexpect import *
from ptyprocess import PtyProcess

from cpython.ref cimport Py_INCREF
from libc.signal cimport *
from posix.signal cimport killpg
from posix.unistd cimport getpid, getpgid, fork

from time import sleep

from sage.cpython.string cimport str_to_bytes
from sage.interfaces.process cimport ContainChildren


class SageSpawn(spawn):
    def __init__(self, *args, **kwds):
        """
        Spawn a subprocess in a pseudo-tty.

        - ``*args``, ``**kwds`` -- see :class:`pexpect.spawn`

        - ``name`` -- human-readable name for this process, used for
          display purposes only

        - ``quit_string`` -- (default: ``None``) if not ``None``, send
          this string to the child process before killing it

        EXAMPLES::

            sage: from sage.interfaces.sagespawn import SageSpawn
            sage: SageSpawn("sleep 1", name="Sleeping Beauty")
            Sleeping Beauty with PID ... running ...
        """
        self.__name = kwds.pop("name", self.__class__.__name__)
        self.quit_string = kwds.pop("quit_string", None)

        kwds.setdefault("ignore_sighup", True)

        # Use a *serious* read buffer of 4MiB, not the ridiculous 2000 bytes
        # that pexpect uses by default.
        kwds.setdefault("maxread", 4194304)

        with ContainChildren(silent=True):
            spawn.__init__(self, *args, **kwds)

        self.delaybeforesend = None
        self.delayafterread = None

    def _spawnpty(self, args, **kwds):
        """
        Create an instance of :class:`SagePtyProcess`.

        EXAMPLES::

            sage: from sage.interfaces.sagespawn import SageSpawn
            sage: s = SageSpawn("sleep 1")
            sage: s.ptyproc
            SagePtyProcess.spawn(...)
        """
        ptyproc = SagePtyProcess.spawn(args, **kwds)
        ptyproc.quit_string = self.quit_string
        return ptyproc

    def __repr__(self):
        """
        A string representing this subprocess.

        EXAMPLES::

            sage: from sage.interfaces.sagespawn import SageSpawn
            sage: s = SageSpawn("true", name="stupid process")
            sage: s  # indirect doctest
            stupid process with PID ... running .../true
            sage: while s.isalive():  # Wait until the process finishes
            ....:     sleep(float(0.1))
            sage: s  # indirect doctest
            stupid process finished running .../true
        """
        try:
            cmd = " ".join(self.args)
            if not self.terminated:
                return "%s with PID %s running %s" % (self.__name, self.pid, cmd)
            else:
                return "%s finished running %s" % (self.__name, cmd)
        except Exception:
            return object.__repr__(self)

    def _keep_alive(self):
        """
        Ensure that the child process does not get terminated, even if
        all references to ``self`` are deleted.

        This is needed in forked Sage processes: a forked process
        should call this method and then delete all references to
        ``self``.

        EXAMPLES::

            sage: from sage.interfaces.sagespawn import SageSpawn
            sage: s = SageSpawn("sh", ["-c", "while true; do sleep 1; done"])
            sage: s._keep_alive()
            sage: pid = s.pid
            sage: del s
            sage: import gc
            sage: _ = gc.collect()

        All references to ``s`` have been deleted, but the process is
        still running::

            sage: from signal import SIGTERM
            sage: os.kill(pid, SIGTERM)
        """
        Py_INCREF(self)

    def expect_peek(self, *args, **kwds):
        r"""
        Like :meth:`expect` but restore the read buffer such that it
        looks like nothing was actually read. The next reading will
        continue at the current position.

        EXAMPLES::

            sage: from sage.interfaces.sagespawn import SageSpawn
            sage: E = SageSpawn("sh", ["-c", "echo hello world"])
            sage: _ = E.expect_peek("w")
            sage: E.read().decode('ascii')
            'hello world\r\n'
        """
        ret = self.expect(*args, **kwds)
        self._before = self.buffer_type()
        self._before.write(self.before + self.after + self.buffer)
        return ret

    def expect_upto(self, *args, **kwds):
        r"""
        Like :meth:`expect` but restore the read buffer starting from
        the matched string. The next reading will continue starting
        with the matched string.

        EXAMPLES::

            sage: from sage.interfaces.sagespawn import SageSpawn
            sage: E = SageSpawn("sh", ["-c", "echo hello world"])
            sage: _ = E.expect_upto("w")
            sage: E.read().decode('ascii')
            'world\r\n'
        """
        ret = self.expect(*args, **kwds)
        self._before = self.buffer_type()
        self._before.write(self.after + self.buffer)
        return ret


class SagePtyProcess(PtyProcess):
    def close(self, force=None):
        """
        Quit the child process: send the quit string, close the
        pseudo-tty and kill the process.

        This function returns immediately, it doesn't wait for the
        child process to die.

        EXAMPLES::

            sage: from sage.interfaces.sagespawn import SageSpawn
            sage: s = SageSpawn("sleep 1000")
            sage: s.close()
            sage: while s.isalive():  # long time (5 seconds)
            ....:     sleep(float(0.1))
        """
        if not self.closed:
            if self.quit_string is not None:
                try:
                    # This can fail if the process already exited
                    # PtyProcess.write takes bytes; ideally we would use
                    # an encoding picked specifically for the target process
                    # but the default (UTF-8) will do now, since I don't
                    # think we have any non-ASCII quit_strings anyways.
                    self.write(str_to_bytes(self.quit_string))
                except (OSError, IOError):
                    pass
            self.fileobj.close()
            self.fd = -1
            self.closed = 1
            self.terminate_async()

    def terminate_async(self, interval=5.0):
        """
        Terminate the child process group asynchronously.

        This function returns immediately, while the child is slowly
        being killed in the background.

        INPUT:

        - ``interval`` -- (default: 5) how much seconds to wait between
          sending two signals

        EXAMPLES:

        Run an infinite loop in the shell::

            sage: from sage.interfaces.sagespawn import SageSpawn
            sage: s = SageSpawn("sh", ["-c", "while true; do sleep 1; done"])

        Check that the process eventually dies after calling
        ``terminate_async``::

            sage: s.ptyproc.terminate_async(interval=float(0.2))
            sage: while True:
            ....:     try:
            ....:         os.kill(s.pid, 0)
            ....:     except OSError:
            ....:         sleep(float(0.1))
            ....:     else:
            ....:         break  # process got killed
        """
        cdef int pg, counter = 0
        cdef int thispg = getpgid(getpid())
        assert thispg != -1

        with ContainChildren():
            if fork():
                # Parent process
                return

            # We need to avoid a race condition where the spawned
            # process has not started up completely yet: we need to
            # wait until the spawned process has changed its process
            # group. See Issue #18741.
            pg = getpgid(self.pid)
            while pg == thispg:
                counter += 1
                if counter >= 20:
                    # Something is seriously wrong, give up...
                    raise RuntimeError("%r is not starting up" % self)
                sleep(interval * 0.125)
                pg = getpgid(self.pid)

            # If we failed to determine the process group, probably
            # this child is already dead...
            if pg == -1:
                return

            # If any of these killpg() calls fail, it's most likely
            # because the process is actually killed.
            if killpg(pg, SIGCONT):
                return
            sleep(interval)
            if killpg(pg, SIGINT):
                return
            sleep(interval)
            if killpg(pg, SIGHUP):
                return
            sleep(interval)
            if killpg(pg, SIGTERM):
                return
            sleep(interval)
            if killpg(pg, SIGKILL):
                return



# sage_setup: distribution = sagemath-categories
"""
Abstract base classes for rings
"""
from sage.rings.ring import IntegralDomain


class NumberField_quadratic(Field):
    r"""
    Abstract base class for :class:`~sage.rings.number_field.number_field.NumberField_quadratic`.

    This class is defined for the purpose of ``isinstance`` tests.  It should not be
    instantiated.

    EXAMPLES::

        sage: import sage.rings.abc
        sage: K.<sqrt2> = QuadraticField(2)                                             # needs sage.rings.number_field
        sage: isinstance(K, sage.rings.abc.NumberField_quadratic)                       # needs sage.rings.number_field
        True

    By design, there is a unique direct subclass::

        sage: sage.rings.abc.NumberField_quadratic.__subclasses__()                     # needs sage.rings.number_field
        [<class 'sage.rings.number_field.number_field.NumberField_quadratic'>]

        sage: len(sage.rings.abc.NumberField_quadratic.__subclasses__()) <= 1
        True
    """

    pass


class NumberField_cyclotomic(Field):
    r"""
    Abstract base class for :class:`~sage.rings.number_field.number_field.NumberField_cyclotomic`.

    This class is defined for the purpose of ``isinstance`` tests.  It should not be
    instantiated.

    EXAMPLES::

        sage: import sage.rings.abc
        sage: K.<zeta> = CyclotomicField(15)                                            # needs sage.rings.number_field
        sage: isinstance(K, sage.rings.abc.NumberField_cyclotomic)                      # needs sage.rings.number_field
        True

    By design, there is a unique direct subclass::

        sage: sage.rings.abc.NumberField_cyclotomic.__subclasses__()                    # needs sage.rings.number_field
        [<class 'sage.rings.number_field.number_field.NumberField_cyclotomic'>]

        sage: len(sage.rings.abc.NumberField_cyclotomic.__subclasses__()) <= 1
        True
    """

    pass


class UniversalCyclotomicField(Field):
    r"""
    Abstract base class for :class:`~sage.rings.universal_cyclotomic_field.UniversalCyclotomicField`.

    This class is defined for the purpose of :func:`isinstance` tests.  It should not be
    instantiated.

    EXAMPLES::

        sage: import sage.rings.abc
        sage: K = UniversalCyclotomicField()                                            # needs sage.libs.gap sage.rings.number_field
        sage: isinstance(K, sage.rings.abc.UniversalCyclotomicField)                    # needs sage.libs.gap sage.rings.number_field
        True

    By design, there is a unique direct subclass::

        sage: sage.rings.abc.UniversalCyclotomicField.__subclasses__()                  # needs sage.libs.gap sage.rings.number_field
        [<class 'sage.rings.universal_cyclotomic_field.UniversalCyclotomicField'>]

        sage: len(sage.rings.abc.NumberField_cyclotomic.__subclasses__()) <= 1
        True
    """

    pass


class AlgebraicField_common(Field):
    r"""
    Abstract base class for :class:`~sage.rings.qqbar.AlgebraicField_common`.

    This class is defined for the purpose of ``isinstance`` tests.  It should not be
    instantiated.

    EXAMPLES::

        sage: import sage.rings.abc
        sage: isinstance(QQbar, sage.rings.abc.AlgebraicField_common)                   # needs sage.rings.number_field
        True
        sage: isinstance(AA, sage.rings.abc.AlgebraicField_common)                      # needs sage.rings.number_field
        True

    By design, other than the abstract subclasses :class:`~sage.rings.abc.AlgebraicField`
    and :class:`~sage.rings.abc.AlgebraicRealField`, there is only one direct implementation
    subclass::

        sage: sage.rings.abc.AlgebraicField_common.__subclasses__()                     # needs sage.rings.number_field
        [<class 'sage.rings.abc.AlgebraicField'>,
         <class 'sage.rings.abc.AlgebraicRealField'>,
         <class 'sage.rings.qqbar.AlgebraicField_common'>]

        sage: len(sage.rings.abc.AlgebraicField_common.__subclasses__()) <= 3
        True
    """

    pass


class AlgebraicField(AlgebraicField_common):
    r"""
    Abstract base class for :class:`~sage.rings.qqbar.AlgebraicField`.

    This class is defined for the purpose of ``isinstance`` tests.  It should not be
    instantiated.

    EXAMPLES::

        sage: import sage.rings.abc
        sage: isinstance(QQbar, sage.rings.abc.AlgebraicField)                          # needs sage.rings.number_field
        True
        sage: isinstance(AA, sage.rings.abc.AlgebraicField)                             # needs sage.rings.number_field
        False

    By design, there is a unique direct subclass::

        sage: sage.rings.abc.AlgebraicField.__subclasses__()                            # needs sage.rings.number_field
        [<class 'sage.rings.qqbar.AlgebraicField'>]

        sage: len(sage.rings.abc.AlgebraicField.__subclasses__()) <= 1
        True
    """

    pass


class AlgebraicRealField(AlgebraicField_common):
    r"""
    Abstract base class for :class:`~sage.rings.qqbar.AlgebraicRealField`.

    This class is defined for the purpose of ``isinstance`` tests.  It should not be
    instantiated.

    EXAMPLES::

        sage: import sage.rings.abc
        sage: isinstance(QQbar, sage.rings.abc.AlgebraicRealField)                      # needs sage.rings.number_field
        False
        sage: isinstance(AA, sage.rings.abc.AlgebraicRealField)                         # needs sage.rings.number_field
        True

    By design, there is a unique direct subclass::

        sage: sage.rings.abc.AlgebraicRealField.__subclasses__()                        # needs sage.rings.number_field
        [<class 'sage.rings.qqbar.AlgebraicRealField'>]

        sage: len(sage.rings.abc.AlgebraicRealField.__subclasses__()) <= 1
        True
    """

    pass


cdef class RealField(Field):
    r"""
    Abstract base class for :class:`~sage.rings.real_mpfr.RealField_class`.

    This class is defined for the purpose of ``isinstance`` tests.  It should not be
    instantiated.

    EXAMPLES::

        sage: import sage.rings.abc
        sage: isinstance(RR, sage.rings.abc.RealField)                                  # needs sage.rings.real_mpfr
        True

    By design, there is a unique direct subclass::

        sage: sage.rings.abc.RealField.__subclasses__()                                 # needs sage.rings.real_mpfr
        [<class 'sage.rings.real_mpfr.RealField_class'>]

        sage: len(sage.rings.abc.RealField.__subclasses__()) <= 1
        True
    """

    pass


class RealBallField(Field):
    r"""
    Abstract base class for :class:`~sage.rings.real_arb.RealBallField`.

    This class is defined for the purpose of ``isinstance`` tests.  It should not be
    instantiated.

    EXAMPLES::

        sage: import sage.rings.abc
        sage: isinstance(RBF, sage.rings.abc.RealBallField)                             # needs sage.libs.flint
        True

    By design, there is a unique direct subclass::

        sage: sage.rings.abc.RealBallField.__subclasses__()                             # needs sage.libs.flint
        [<class 'sage.rings.real_arb.RealBallField'>]

        sage: len(sage.rings.abc.RealBallField.__subclasses__()) <= 1
        True
    """

    pass


cdef class RealIntervalField(Field):
    r"""
    Abstract base class for :class:`~sage.rings.real_mpfi.RealIntervalField_class`.

    This class is defined for the purpose of ``isinstance`` tests.  It should not be
    instantiated.

    EXAMPLES::

        sage: import sage.rings.abc
        sage: isinstance(RIF, sage.rings.abc.RealIntervalField)                         # needs sage.rings.real_interval_field
        True

    By design, there is a unique direct subclass::

        sage: sage.rings.abc.RealIntervalField.__subclasses__()                         # needs sage.rings.real_interval_field
        [<class 'sage.rings.real_mpfi.RealIntervalField_class'>]

        sage: len(sage.rings.abc.RealIntervalField.__subclasses__()) <= 1
        True
    """

    pass


cdef class RealDoubleField(Field):
    r"""
    Abstract base class for :class:`~sage.rings.real_double.RealDoubleField_class`.

    This class is defined for the purpose of ``isinstance`` tests.  It should not be
    instantiated.

    EXAMPLES::

        sage: import sage.rings.abc
        sage: isinstance(RDF, sage.rings.abc.RealDoubleField)
        True

    By design, there is a unique direct subclass::

        sage: sage.rings.abc.RealDoubleField.__subclasses__()
        [<class 'sage.rings.real_double.RealDoubleField_class'>]

        sage: len(sage.rings.abc.RealDoubleField.__subclasses__()) <= 1
        True
    """

    pass


cdef class ComplexField(Field):
    r"""
    Abstract base class for :class:`~sage.rings.complex_mpfr.ComplexField_class`.

    This class is defined for the purpose of ``isinstance`` tests.  It should not be
    instantiated.

    EXAMPLES::

        sage: import sage.rings.abc
        sage: isinstance(CC, sage.rings.abc.ComplexField)                               # needs sage.rings.real_mpfr
        True

    By design, there is a unique direct subclass::

        sage: sage.rings.abc.ComplexField.__subclasses__()                              # needs sage.rings.real_mpfr
        [<class 'sage.rings.complex_mpfr.ComplexField_class'>]

        sage: len(sage.rings.abc.ComplexField.__subclasses__()) <= 1
        True
    """

    pass


class ComplexBallField(Field):
    r"""
    Abstract base class for :class:`~sage.rings.complex_arb.ComplexBallField`.

    This class is defined for the purpose of ``isinstance`` tests.  It should not be
    instantiated.

    EXAMPLES::

        sage: import sage.rings.abc
        sage: isinstance(CBF, sage.rings.abc.ComplexBallField)                          # needs sage.libs.flint
        True

    By design, there is a unique direct subclass::

        sage: sage.rings.abc.ComplexBallField.__subclasses__()                          # needs sage.libs.flint
        [<class 'sage.rings.complex_arb.ComplexBallField'>]

        sage: len(sage.rings.abc.ComplexBallField.__subclasses__()) <= 1
        True
    """

    pass


class ComplexIntervalField(Field):
    r"""
    Abstract base class for :class:`~sage.rings.complex_interval_field.ComplexIntervalField_class`.

    This class is defined for the purpose of ``isinstance`` tests.  It should not be
    instantiated.

    EXAMPLES::

        sage: import sage.rings.abc
        sage: isinstance(CIF, sage.rings.abc.ComplexIntervalField)                      # needs sage.rings.complex_interval_field
        True

    By design, there is a unique direct subclass::

        sage: sage.rings.abc.ComplexIntervalField.__subclasses__()                      # needs sage.rings.complex_interval_field
        [<class 'sage.rings.complex_interval_field.ComplexIntervalField_class'>]

        sage: len(sage.rings.abc.ComplexIntervalField.__subclasses__()) <= 1
        True
    """

    pass


cdef class ComplexDoubleField(Field):
    r"""
    Abstract base class for :class:`~sage.rings.complex_double.ComplexDoubleField_class`.

    This class is defined for the purpose of ``isinstance`` tests.  It should not be
    instantiated.

    EXAMPLES::

        sage: import sage.rings.abc
        sage: isinstance(CDF, sage.rings.abc.ComplexDoubleField)                        # needs sage.rings.complex_double
        True

    By design, there is a unique direct subclass::

        sage: sage.rings.abc.ComplexDoubleField.__subclasses__()                        # needs sage.rings.complex_double
        [<class 'sage.rings.complex_double.ComplexDoubleField_class'>]

        sage: len(sage.rings.abc.ComplexDoubleField.__subclasses__()) <= 1
        True
    """

    pass


class IntegerModRing:
    r"""
    Abstract base class for :class:`~sage.rings.finite_rings.integer_mod_ring.IntegerModRing_generic`.

    This class is defined for the purpose of ``isinstance`` tests.  It should not be
    instantiated.

    EXAMPLES::

        sage: import sage.rings.abc
        sage: isinstance(Integers(7), sage.rings.abc.IntegerModRing)
        True

    By design, there is a unique direct subclass::

        sage: sage.rings.abc.IntegerModRing.__subclasses__()
        [<class 'sage.rings.finite_rings.integer_mod_ring.IntegerModRing_generic'>]

        sage: len(sage.rings.abc.IntegerModRing.__subclasses__()) <= 1
        True
    """

    pass


class Order:
    r"""
    Abstract base class for :class:`~sage.rings.number_field.order.Order`.

    This class is defined for the purpose of ``isinstance`` tests.  It should not be
    instantiated.

    EXAMPLES::

        sage: import sage.rings.abc
        sage: x = polygen(ZZ, 'x')
        sage: K.<a> = NumberField(x^2 + 1); O = K.order(2*a)                            # needs sage.rings.number_field
        sage: isinstance(O, sage.rings.abc.Order)                                       # needs sage.rings.number_field
        True

    By design, there is a unique direct subclass::

        sage: sage.rings.abc.Order.__subclasses__()                                     # needs sage.rings.number_field
        [<class 'sage.rings.number_field.order.Order'>]

        sage: len(sage.rings.abc.Order.__subclasses__()) <= 1
        True
    """

    pass


class pAdicRing(IntegralDomain):
    r"""
    Abstract base class for :class:`~sage.rings.padics.generic_nodes.pAdicRingGeneric`.

    This class is defined for the purpose of ``isinstance`` tests.  It should not be
    instantiated.

    EXAMPLES::

        sage: import sage.rings.abc
        sage: isinstance(Zp(5), sage.rings.abc.pAdicRing)                               # needs sage.rings.padics
        True
        sage: isinstance(Qp(5), sage.rings.abc.pAdicRing)                               # needs sage.rings.padics
        False

    By design, there is a unique direct subclass::

        sage: sage.rings.abc.pAdicRing.__subclasses__()                                 # needs sage.rings.padics
        [<class 'sage.rings.padics.generic_nodes.pAdicRingGeneric'>]

        sage: len(sage.rings.abc.pAdicRing.__subclasses__()) <= 1
        True
    """

    pass


class pAdicField(Field):
    r"""
    Abstract base class for :class:`~sage.rings.padics.generic_nodes.pAdicFieldGeneric`.

    This class is defined for the purpose of ``isinstance`` tests.  It should not be
    instantiated.

    EXAMPLES::

        sage: import sage.rings.abc
        sage: isinstance(Zp(5), sage.rings.abc.pAdicField)                              # needs sage.rings.padics
        False
        sage: isinstance(Qp(5), sage.rings.abc.pAdicField)                              # needs sage.rings.padics
        True

    By design, there is a unique direct subclass::

        sage: sage.rings.abc.pAdicField.__subclasses__()                                # needs sage.rings.padics
        [<class 'sage.rings.padics.generic_nodes.pAdicFieldGeneric'>]

        sage: len(sage.rings.abc.pAdicField.__subclasses__()) <= 1
        True
    """

    pass


cdef class SymbolicRing(CommutativeRing):
    r"""
    Abstract base class for :class:`~sage.rings.symbolic.ring.SymbolicRing`.

    This class is defined for the purpose of ``isinstance`` tests.  It should not be
    instantiated.

    EXAMPLES::

        sage: import sage.rings.abc
        sage: isinstance(SR, sage.rings.abc.SymbolicRing)                               # needs sage.symbolic
        True

    By design, other than the abstract subclass :class:`~sage.rings.abc.CallableSymbolicExpressionRing`,
    there is only one direct implementation subclass::

        sage: sage.rings.abc.SymbolicRing.__subclasses__()                              # needs sage.symbolic
        [<class 'sage.rings.abc.CallableSymbolicExpressionRing'>,
         <class 'sage.symbolic.ring.SymbolicRing'>]

        sage: len(sage.rings.abc.SymbolicRing.__subclasses__()) <= 2
        True
    """

    pass


class CallableSymbolicExpressionRing(SymbolicRing):
    r"""
    Abstract base class for :class:`~sage.rings.symbolic.callable.CallableSymbolicExpressionRing_class`.

    This class is defined for the purpose of ``isinstance`` tests.  It should not be
    instantiated.

    EXAMPLES::

        sage: import sage.rings.abc
        sage: f = x.function(x).parent()                                                # needs sage.symbolic
        sage: isinstance(f, sage.rings.abc.CallableSymbolicExpressionRing)              # needs sage.symbolic
        True

    By design, there is a unique direct subclass::

        sage: sage.rings.abc.CallableSymbolicExpressionRing.__subclasses__()            # needs sage.symbolic
        [<class 'sage.symbolic.callable.CallableSymbolicExpressionRing_class'>]

        sage: len(sage.rings.abc.CallableSymbolicExpressionRing.__subclasses__()) <= 1
        True
    """

    pass



# sage_setup: distribution = sagemath-categories
r"""
Class to flatten polynomial rings over polynomial ring

For example ``QQ['a','b'],['x','y']`` flattens to ``QQ['a','b','x','y']``.

EXAMPLES::

    sage: R = QQ['x']['y']['s','t']['X']
    sage: from sage.rings.polynomial.flatten import FlatteningMorphism
    sage: phi = FlatteningMorphism(R); phi
    Flattening morphism:
      From: Univariate Polynomial Ring in X
            over Multivariate Polynomial Ring in s, t
            over Univariate Polynomial Ring in y
            over Univariate Polynomial Ring in x over Rational Field
      To:   Multivariate Polynomial Ring in x, y, s, t, X over Rational Field
    sage: phi('x*y*s + t*X').parent()
    Multivariate Polynomial Ring in x, y, s, t, X over Rational Field

Authors:

Vincent Delecroix, Ben Hutz (July 2016): initial implementation
"""

# ****************************************************************************
#                  Copyright (C) 2016
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

import itertools

from sage.categories.homset import Homset
from sage.categories.morphism import Morphism
from sage.misc.cachefunc import cached_method
from .polynomial_ring_constructor import PolynomialRing
from .polynomial_ring import PolynomialRing_general
from .multi_polynomial_ring_base import MPolynomialRing_base
from sage.rings.fraction_field import FractionField_generic
from sage.rings.fraction_field_element import FractionFieldElement
from sage.rings.polynomial.polydict import ETuple


class FlatteningMorphism(Morphism):
    r"""
    EXAMPLES::

        sage: R = QQ['a','b']['x','y','z']['t1','t2']
        sage: from sage.rings.polynomial.flatten import FlatteningMorphism
        sage: f = FlatteningMorphism(R)
        sage: f.codomain()
        Multivariate Polynomial Ring in a, b, x, y, z, t1, t2 over Rational Field
        sage: p = R('(a+b)*x + (a^2-b)*t2*(z+y)')
        sage: p
        ((a^2 - b)*y + (a^2 - b)*z)*t2 + (a + b)*x
        sage: f(p)
        a^2*y*t2 + a^2*z*t2 - b*y*t2 - b*z*t2 + a*x + b*x
        sage: f(p).parent()
        Multivariate Polynomial Ring in a, b, x, y, z, t1, t2 over Rational Field

    Also works when univariate polynomial ring are involved::

        sage: R = QQ['x']['y']['s','t']['X']
        sage: from sage.rings.polynomial.flatten import FlatteningMorphism
        sage: f = FlatteningMorphism(R)
        sage: f.codomain()
        Multivariate Polynomial Ring in x, y, s, t, X over Rational Field
        sage: p = R('((x^2 + 1) + (x+2)*y + x*y^3)*(s+t) + x*y*X')
        sage: p
        x*y*X + (x*y^3 + (x + 2)*y + x^2 + 1)*s + (x*y^3 + (x + 2)*y + x^2 + 1)*t
        sage: f(p)
        x*y^3*s + x*y^3*t + x^2*s + x*y*s + x^2*t + x*y*t + x*y*X + 2*y*s + 2*y*t + s + t
        sage: f(p).parent()
        Multivariate Polynomial Ring in x, y, s, t, X over Rational Field
    """
    def __init__(self, domain):
        """
        The Python constructor.

        EXAMPLES::

            sage: R = ZZ['a', 'b', 'c']['x', 'y', 'z']
            sage: from sage.rings.polynomial.flatten import FlatteningMorphism
            sage: FlatteningMorphism(R)
            Flattening morphism:
              From: Multivariate Polynomial Ring in x, y, z over Multivariate Polynomial Ring in a, b, c over Integer Ring
              To:   Multivariate Polynomial Ring in a, b, c, x, y, z over Integer Ring

        ::

            sage: R = ZZ['a']['b']['c']
            sage: from sage.rings.polynomial.flatten import FlatteningMorphism
            sage: FlatteningMorphism(R)
            Flattening morphism:
              From: Univariate Polynomial Ring in c over Univariate Polynomial Ring in b over Univariate Polynomial Ring in a over Integer Ring
              To:   Multivariate Polynomial Ring in a, b, c over Integer Ring

        ::

            sage: R = ZZ['a']['a','b']
            sage: from sage.rings.polynomial.flatten import FlatteningMorphism
            sage: FlatteningMorphism(R)
            Flattening morphism:
              From: Multivariate Polynomial Ring in a, b over Univariate Polynomial Ring in a over Integer Ring
              To:   Multivariate Polynomial Ring in a, a0, b over Integer Ring

        ::

            sage: # needs sage.rings.number_field
            sage: x = polygen(ZZ, 'x')
            sage: K.<v> = NumberField(x^3 - 2)
            sage: R = K['x','y']['a','b']
            sage: from sage.rings.polynomial.flatten import FlatteningMorphism
            sage: f = FlatteningMorphism(R)
            sage: f(R('v*a*x^2 + b^2 + 1/v*y'))
            v*x^2*a + b^2 + (1/2*v^2)*y

        ::

            sage: # needs sage.rings.number_field
            sage: R = QQbar['x','y']['a','b']
            sage: from sage.rings.polynomial.flatten import FlatteningMorphism
            sage: f = FlatteningMorphism(R)
            sage: f(R('QQbar(sqrt(2))*a*x^2 + b^2 + QQbar(I)*y'))                       # needs sage.symbolic
            1.414213562373095?*x^2*a + b^2 + I*y

        ::

            sage: # needs sage.rings.number_field
            sage: R.<z> = PolynomialRing(QQbar, 1)
            sage: from sage.rings.polynomial.flatten import FlatteningMorphism
            sage: f = FlatteningMorphism(R)
            sage: f.domain(), f.codomain()
            (Multivariate Polynomial Ring in z over Algebraic Field,
             Multivariate Polynomial Ring in z over Algebraic Field)

        ::

            sage: # needs sage.rings.number_field
            sage: R.<z> = PolynomialRing(QQbar)
            sage: from sage.rings.polynomial.flatten import FlatteningMorphism
            sage: f = FlatteningMorphism(R)
            sage: f.domain(), f.codomain()
            (Univariate Polynomial Ring in z over Algebraic Field,
             Univariate Polynomial Ring in z over Algebraic Field)

        TESTS::

            sage: Pol = QQ['x']['x0']['x']
            sage: fl = FlatteningMorphism(Pol)
            sage: fl
            Flattening morphism:
              From: Univariate Polynomial Ring in x over Univariate Polynomial Ring in x0 over Univariate Polynomial Ring in x over Rational Field
              To:   Multivariate Polynomial Ring in x, x0, x1 over Rational Field
            sage: p = Pol([[[1,2],[3,4]],[[5,6],[7,8]]])
            sage: fl.section()(fl(p)) == p
            True
        """
        if not isinstance(domain, PolynomialRing_general) and not isinstance(domain, MPolynomialRing_base):
            raise ValueError("domain should be a polynomial ring")

        ring = domain
        variables = []
        intermediate_rings = []

        while isinstance(ring, PolynomialRing_general) or isinstance(ring, MPolynomialRing_base):
            intermediate_rings.append(ring)
            v = ring.variable_names()
            variables.extend(reversed(v))
            ring = ring.base_ring()
        self._intermediate_rings = intermediate_rings
        variables.reverse()
        for i, a in enumerate(variables):
            if a in variables[:i]:
                for index in itertools.count():
                    b = a + str(index)
                    if b not in variables:  # not just variables[:i]!
                        break
                variables[i] = b
        if isinstance(domain, MPolynomialRing_base):
            codomain = PolynomialRing(ring, variables, len(variables))
        else:
            codomain = PolynomialRing(ring, variables)

        hom = Homset(domain, codomain, base=ring, check=False)
        Morphism.__init__(self, hom)
        self._repr_type_str = 'Flattening'

    def _call_(self, p):
        r"""
        Evaluate a flattening morphism.

        EXAMPLES::

            sage: R = QQ['a','b','c']['x','y','z']
            sage: from sage.rings.polynomial.flatten import FlatteningMorphism
            sage: h = FlatteningMorphism(R)('2*a*x + b*z'); h
            2*a*x + b*z
            sage: h.parent()
            Multivariate Polynomial Ring in a, b, c, x, y, z over Rational Field

        TESTS::

            sage: R = QQ['x']['y']['s','t']
            sage: p = R('s*x + y*t + x^2*s + 1 + t')
            sage: from sage.rings.polynomial.flatten import FlatteningMorphism
            sage: f = FlatteningMorphism(R)
            sage: f._call_(p)
            x^2*s + x*s + y*t + t + 1
        """
        # If we are just specializing a univariate polynomial, then
        # the flattening morphism is the identity
        if self.codomain().ngens() == 1:
            return p

        p = {(): p}

        for ring in self._intermediate_rings:
            new_p = {}
            if isinstance(ring, PolynomialRing_general):
                for mon, pp in p.items():
                    assert pp.parent() is ring
                    for i, j in pp.monomial_coefficients().items():
                        new_p[(i,) + (mon)] = j
            elif isinstance(ring, MPolynomialRing_base):
                for mon, pp in p.items():
                    assert pp.parent() is ring
                    for mmon, q in pp.monomial_coefficients().items():
                        new_p[tuple(mmon) + mon] = q
            else:
                raise RuntimeError
            p = new_p

        return self.codomain()(p, check=False)

    @cached_method
    def section(self):
        """
        Inverse of this flattening morphism.

        EXAMPLES::

            sage: R = QQ['a','b','c']['x','y','z']
            sage: from sage.rings.polynomial.flatten import FlatteningMorphism
            sage: h = FlatteningMorphism(R)
            sage: h.section()
            Unflattening morphism:
              From: Multivariate Polynomial Ring in a, b, c, x, y, z over Rational Field
              To:   Multivariate Polynomial Ring in x, y, z
                    over Multivariate Polynomial Ring in a, b, c over Rational Field

        ::

            sage: R = ZZ['a']['b']['c']
            sage: from sage.rings.polynomial.flatten import FlatteningMorphism
            sage: FlatteningMorphism(R).section()
            Unflattening morphism:
              From: Multivariate Polynomial Ring in a, b, c over Integer Ring
              To:   Univariate Polynomial Ring in c over Univariate Polynomial Ring in b
                    over Univariate Polynomial Ring in a over Integer Ring
        """
        return UnflatteningMorphism(self.codomain(), self.domain())

    def inverse(self):
        """
        Return the inverse of this flattening morphism.

        This is the same as calling :meth:`section`.

        EXAMPLES::

            sage: f = QQ['x,y']['u,v'].flattening_morphism()
            sage: f.inverse()
            Unflattening morphism:
              From: Multivariate Polynomial Ring in x, y, u, v over Rational Field
              To:   Multivariate Polynomial Ring in u, v
                    over Multivariate Polynomial Ring in x, y over Rational Field
        """
        return self.section()


class UnflatteningMorphism(Morphism):
    r"""
    Inverses for :class:`FlatteningMorphism`.

    EXAMPLES::

        sage: R = QQ['c','x','y','z']
        sage: S = QQ['c']['x','y','z']
        sage: from sage.rings.polynomial.flatten import UnflatteningMorphism
        sage: f = UnflatteningMorphism(R, S)
        sage: g = f(R('x^2 + c*y^2 - z^2'));g
        x^2 + c*y^2 - z^2
        sage: g.parent()
        Multivariate Polynomial Ring in x, y, z
         over Univariate Polynomial Ring in c over Rational Field

    ::

        sage: R = QQ['a','b', 'x','y']
        sage: S = QQ['a','b']['x','y']
        sage: from sage.rings.polynomial.flatten import UnflatteningMorphism
        sage: UnflatteningMorphism(R, S)
        Unflattening morphism:
          From: Multivariate Polynomial Ring in a, b, x, y over Rational Field
          To:   Multivariate Polynomial Ring in x, y
                over Multivariate Polynomial Ring in a, b over Rational Field
    """

    def __init__(self, domain, codomain):
        """
        The Python constructor.

        EXAMPLES::

            sage: R = QQ['x']['y']['s','t']['X']
            sage: p = R.random_element()
            sage: from sage.rings.polynomial.flatten import FlatteningMorphism
            sage: f = FlatteningMorphism(R)
            sage: g = f.section()
            sage: g(f(p)) == p
            True

        ::

            sage: R = QQ['a','b','x','y']
            sage: S = ZZ['a','b']['x','z']
            sage: from sage.rings.polynomial.flatten import UnflatteningMorphism
            sage: UnflatteningMorphism(R, S)
            Traceback (most recent call last):
            ...
            ValueError: rings must have same base ring

        ::

            sage: R = QQ['a','b','x','y']
            sage: S = QQ['a','b']['x','z','w']
            sage: from sage.rings.polynomial.flatten import UnflatteningMorphism
            sage: UnflatteningMorphism(R, S)
            Traceback (most recent call last):
            ...
            ValueError: rings must have the same number of variables
        """
        if not isinstance(domain, MPolynomialRing_base):
            raise ValueError("domain should be a multivariate polynomial ring")
        if not isinstance(codomain, PolynomialRing_general) and not isinstance(codomain, MPolynomialRing_base):
            raise ValueError("codomain should be a polynomial ring")

        ring = codomain
        intermediate_rings = []

        while True:
            is_polynomial_ring = isinstance(ring, PolynomialRing_general)
            if not (is_polynomial_ring or isinstance(ring, MPolynomialRing_base)):
                break
            intermediate_rings.append((ring, is_polynomial_ring))
            ring = ring.base_ring()

        if domain.base_ring() != intermediate_rings[-1][0].base_ring():
            raise ValueError("rings must have same base ring")
        if domain.ngens() != sum([R.ngens() for R, _ in intermediate_rings]):
            raise ValueError("rings must have the same number of variables")

        self._intermediate_rings = intermediate_rings

        hom = Homset(domain, codomain, base=ring, check=False)
        Morphism.__init__(self, hom)
        self._repr_type_str = 'Unflattening'

    def _call_(self, p):
        """
        Evaluate an unflattening morphism.

        TESTS::

            sage: from sage.rings.polynomial.flatten import FlatteningMorphism
            sage: rings = [ZZ['x']['y']['a,b,c']]
            sage: rings += [GF(4)['x','y']['a','b']]                                    # needs sage.rings.finite_rings
            sage: rings += [AA['x']['a','b']['y'], QQbar['a1','a2']['t']['X','Y']]      # needs sage.rings.number_field
            sage: for R in rings:                                                       # needs sage.modules
            ....:    f = FlatteningMorphism(R)
            ....:    g = f.section()
            ....:    for _ in range(10):
            ....:        p = R.random_element()
            ....:        assert p == g(f(p))
            ....:        z = R.zero()
            ....:        assert z == g(f(z))
        """
        index = [0]
        for R, _ in reversed(self._intermediate_rings):
            index.append(index[-1] + len(R.gens()))
        newpol = [{} for _ in self._intermediate_rings]
        expo = sorted(p.exponents(), key=lambda e: tuple(reversed(e)))
        for i in range(len(expo)):
            cur_exp = expo[i]
            for l in range(len(self._intermediate_rings)):
                R, univariate = self._intermediate_rings[-1 - l]
                idx = index[l + 1]
                sub_exp = (cur_exp[index[l]] if univariate
                           else cur_exp[index[l]:idx])
                if l == 0:
                    newpol[l][sub_exp] = p[cur_exp]
                else:
                    newpol[l][sub_exp] = newpol[l - 1]
                    newpol[l - 1] = {}
                if (i == len(expo) - 1 or expo[i + 1][idx:] != cur_exp[idx:]):
                    newpol[l] = R(newpol[l], check=False)
                else:
                    break
        return R(newpol[-1], check=False)


class SpecializationMorphism(Morphism):
    r"""
    Morphisms to specialize parameters in (stacked) polynomial rings.

    EXAMPLES::

        sage: R.<c> = PolynomialRing(QQ)
        sage: S.<x,y,z> = PolynomialRing(R)
        sage: D = dict({c:1})
        sage: from sage.rings.polynomial.flatten import SpecializationMorphism
        sage: f = SpecializationMorphism(S, D)
        sage: g = f(x^2 + c*y^2 - z^2); g
        x^2 + y^2 - z^2
        sage: g.parent()
        Multivariate Polynomial Ring in x, y, z over Rational Field

    ::

        sage: R.<c> = PolynomialRing(QQ)
        sage: S.<z> = PolynomialRing(R)
        sage: from sage.rings.polynomial.flatten import SpecializationMorphism
        sage: xi = SpecializationMorphism(S, {c:0}); xi
        Specialization morphism:
          From: Univariate Polynomial Ring in z
                over Univariate Polynomial Ring in c over Rational Field
          To:   Univariate Polynomial Ring in z over Rational Field
        sage: xi(z^2+c)
        z^2

    ::

        sage: R1.<u,v> = PolynomialRing(QQ)
        sage: R2.<a,b,c> = PolynomialRing(R1)
        sage: S.<x,y,z> = PolynomialRing(R2)
        sage: D = dict({a:1, b:2, x:0, u:1})
        sage: from sage.rings.polynomial.flatten import SpecializationMorphism
        sage: xi = SpecializationMorphism(S, D); xi
        Specialization morphism:
          From: Multivariate Polynomial Ring in x, y, z
                over Multivariate Polynomial Ring in a, b, c
                over Multivariate Polynomial Ring in u, v over Rational Field
          To:   Multivariate Polynomial Ring in y, z over Univariate Polynomial Ring in c
                over Univariate Polynomial Ring in v over Rational Field
        sage: xi(a*(x*z+y^2)*u+b*v*u*(x*z+y^2)*y^2*c+c*y^2*z^2)
        2*v*c*y^4 + c*y^2*z^2 + y^2
    """

    def __init__(self, domain, D):
        """
        The Python constructor.

        EXAMPLES::

            sage: S.<x,y> = PolynomialRing(QQ)
            sage: D = dict({x:1})
            sage: from sage.rings.polynomial.flatten import SpecializationMorphism
            sage: phi = SpecializationMorphism(S, D); phi
            Specialization morphism:
              From: Multivariate Polynomial Ring in x, y over Rational Field
              To:   Univariate Polynomial Ring in y over Rational Field
            sage: phi(x^2 + y^2)
            y^2 + 1

        ::

            sage: R.<a,b,c> = PolynomialRing(ZZ)
            sage: S.<x,y,z> = PolynomialRing(R)
            sage: from sage.rings.polynomial.flatten import SpecializationMorphism
            sage: xi = SpecializationMorphism(S, {a:1/2})
            Traceback (most recent call last):
            ...
            TypeError: no conversion of this rational to integer

        The following was fixed in :issue:`23811`::

            sage: R.<c> = RR[]
            sage: P.<z> = AffineSpace(R, 1)
            sage: H = End(P)
            sage: f = H([z^2 + c])
            sage: f.specialization({c:1})                                               # needs sage.modules
            Scheme endomorphism of
             Affine Space of dimension 1 over Real Field with 53 bits of precision
              Defn: Defined on coordinates by sending (z) to
                    (z^2 + 1.00000000000000)
        """
        if not isinstance(domain, PolynomialRing_general) and not isinstance(domain, MPolynomialRing_base):
            raise TypeError("domain should be a polynomial ring")

        # use only the generators that are in the stack somewhere,
        # and ignore the rest
        all_gens = domain.gens_dict_recursive()
        new_D = {}
        for gen in D:
            if str(gen) in all_gens:
                new_D[gen] = D[gen]
        D = new_D

        # _sub_specialization is a specialization morphism (recursive)
        # which is applied to the base Fraction field, or None if it's
        # any other base ring

        self._sub_specialization = None

        # We use this composition where "flat" is a flattened
        # polynomial ring.
        #
        #            phi       D       psi
        #     domain  →  flat  →  flat  →  R
        #        │         │               │
        #        └─────────┴───────────────┘
        # _flattening_morph     _eval_morph
        #             = phi       = psi ∘ D

        phi = FlatteningMorphism(domain)
        flat = phi.codomain()
        base = flat.base_ring()

        # Change domain of D to "flat" and ensure that the values lie
        # in the base ring.
        D = {phi(k): base(D[k]) for k in D}

        # Construct unflattened codomain R
        new_vars = []
        R = domain
        while isinstance(R, PolynomialRing_general) or isinstance(R, MPolynomialRing_base) or isinstance(R, FractionField_generic):
            if isinstance(R, FractionField_generic):
                # We've hit base_ring, so set _sub_specialization and exit the loop
                field_over = R.base()
                applicable_vars = {key: val for key, val in D.items()
                                   if key not in flat.gens()}
                # If there are any variables in D to set in _sub_specialization
                if applicable_vars:
                    # Coerce the generators to be in the right ring
                    # This un-does changing the domain of D to be in the flat base ring
                    tmp = {}
                    for var, val in applicable_vars.items():
                        for gstr, gen in field_over.gens_dict_recursive().items():
                            if str(var) == gstr:
                                tmp[gen] = val
                                break
                        else:
                            # Should have been caught earlier
                            raise NameError("argument " + str(var) + " is not a generator anywhere in the polynomial tower")
                    applicable_vars = tmp
                    self._sub_specialization = FractionSpecializationMorphism(R, applicable_vars)
                break
            # We're still in the polynomials, so keep track of the tower
            old = R.gens()
            new = [t for t in old if t not in D]
            force_multivariate = ((len(old) == 1) and isinstance(R, MPolynomialRing_base))
            new_vars.append((new, force_multivariate, old))
            R = R.base_ring()

        if self._sub_specialization:
            # The sub_specialization range will be different
            # if it applied some variables from D
            R = self._sub_specialization.codomain().fraction_field()

        # Construct unflattening map psi (only defined on the variables
        # of "flat" which are not involved in D)
        psi = dict()
        # Reconstruct the proper domain of this morphism
        # based on the sub_specialization domains
        new_domain = R
        for new, force_multivariate, old in reversed(new_vars):
            if self._sub_specialization:
                if force_multivariate:
                    new_domain = PolynomialRing(new_domain, old, len(old))
                else:
                    new_domain = PolynomialRing(new_domain, old)
            if not new:
                continue
            var_names = [str(var) for var in new]
            if force_multivariate:
                R = PolynomialRing(R, var_names, len(var_names))
            else:
                R = PolynomialRing(R, var_names)
            # Map variables in "new" to R
            psi.update(zip([phi(w) for w in new], R.gens()))

        # Fix domain of eval_morph
        # (note: phi's domain is correct)
        if self._sub_specialization:
            phi_prime = FlatteningMorphism(new_domain)
            flat_old = flat
            flat = phi_prime.codomain()
            base_prime = flat.base_ring()
            D = {phi(k): base_prime(D[k]) for k in D}
        else:
            # The bottom of our tower has not changed
            def flat_old(x):
                return x

        # Compose D with psi
        vals = []
        for t in flat.gens():
            if t in D:
                vals.append(R.coerce(D[t]))
            else:
                # Make sure keys are in the old domain
                # or else they won't match exactly
                vals.append(psi[flat_old(t)])

        self._flattening_morph = phi
        self._eval_morph = flat.hom(vals, R)
        self._repr_type_str = 'Specialization'
        Morphism.__init__(self, domain, R)

    def _call_(self, p):
        """
        Evaluate a specialization morphism.

        EXAMPLES::

            sage: R.<a,b,c> = PolynomialRing(ZZ)
            sage: S.<x,y,z> = PolynomialRing(R)
            sage: D = dict({a:1, b:2, c:3})
            sage: from sage.rings.polynomial.flatten import SpecializationMorphism
            sage: xi = SpecializationMorphism(S, D)
            sage: xi(a*x + b*y + c*z)
            x + 2*y + 3*z
        """
        flat = self._flattening_morph(p)
        if self._sub_specialization is not None:
            # The base_ring should be a fraction field, so
            # apply _sub_specialization to each coefficient
            # in the flattened polynomial
            tmp = {}
            for exponent, coefficient in flat.monomial_coefficients().items():
                # Fix the type of exponent from (a,) to a
                #     (necessary for R(tmp) later)
                if isinstance(exponent, ETuple) and len(exponent) == 1:
                    exponent = exponent[0]
                # Coefficient should be a fraction
                tmp[exponent] = self._sub_specialization._call_(coefficient)
            # tmp's parent should be the same construction as flat
            # but over _sub_specialization's codomain
            ring_constructor = flat.parent().construction()[0]
            fraction_type = self._sub_specialization.codomain()
            R = ring_constructor(fraction_type)
            flat = R(tmp)
        return self._eval_morph(flat)


class FractionSpecializationMorphism(Morphism):
    """
    A specialization morphism for fraction fields over (stacked) polynomial rings
    """
    def __init__(self, domain, D):
        """
        Initialize the morphism with a domain and dictionary of specializations.

        EXAMPLES::

            sage: R.<a,c> = QQ[]
            sage: S.<x,y> = R[]
            sage: from sage.rings.polynomial.flatten import FractionSpecializationMorphism
            sage: phi = FractionSpecializationMorphism(Frac(S), {c:3})
            sage: phi
            Fraction Specialization morphism:
              From: Fraction Field of Multivariate Polynomial Ring in x, y
                    over Multivariate Polynomial Ring in a, c over Rational Field
              To:   Fraction Field of Multivariate Polynomial Ring in x, y
                    over Univariate Polynomial Ring in a over Rational Field
        """
        if not isinstance(domain, FractionField_generic):
            raise TypeError("domain must be a fraction field")
        self._specialization = SpecializationMorphism(domain.base(), D)
        self._repr_type_str = 'Fraction Specialization'
        Morphism.__init__(self, domain, self._specialization.codomain().fraction_field())

    def _call_(self, p):
        """
        Evaluate a fraction specialization morphism.

        EXAMPLES::

            sage: R.<a,b,c> = QQ[]
            sage: S.<x,y,z> = R[]
            sage: from sage.rings.polynomial.flatten import FractionSpecializationMorphism
            sage: phi = FractionSpecializationMorphism(Frac(S), {a:3, b:2, c:-2})
            sage: spec = phi((a*x + b*y) / (c*z))
            sage: spec
            (3*x + 2*y)/(-2*z)
            sage: spec.parent()
            Fraction Field of Multivariate Polynomial Ring in x, y, z over Rational Field
        """
        if not isinstance(p, FractionFieldElement):
            raise TypeError("p must be a fraction field element")
        numerator = self._specialization._call_(p.numerator())
        denominator = self._specialization._call_(p.denominator())
        return numerator / denominator
